#!/usr/bin/env python3
""" run neurondm related exports and conversions
Usage:
    neurondm-build release [options]
    neurondm-build all [options]
    neurondm-build [indicators phenotypes] [options]
    neurondm-build [models bridge old dep dev] [options]
    neurondm-build [sheets] [options]

Options:
    -h --help                   Display this help message
"""

import io
import os
import re
import csv
import types
from pathlib import Path
from collections import defaultdict
from urllib.parse import quote
import rdflib
import ontquery as oq
from rdflib.extras import infixowl
from pyontutils.core import makeGraph, createOntology, OntId as OntId_, OntConjunctiveGraph
from pyontutils.utils import TODAY, rowParse, refile, makeSimpleLogger, anyMembers
from pyontutils.obo_io import OboFile
from neurondm import _NEURON_CLASS, OntTerm
from neurondm.core import OntTermOntologyOnly, OntTermInterLexOnly, log as _log, auth
from pyontutils.scigraph import Graph, Vocabulary
from pyontutils.namespaces import (makePrefixes,
                                   makeNamespaces,
                                   TEMP,
                                   ilxtr,
                                   BFO,
                                   NIFRID,
                                   definition,
                                   ilx_partOf,
                                   partOf as BFO_partOf,
                                   rdf,
                                   rdfs,
                                   owl)
from itertools import chain
from neurondm.indicators import PhenotypeIndicators

try:
    breakpoint
except NameError:
    from IPython import embed as breakpoint

log = _log.getChild('build')

resources = auth.get_path('resources')

NIFRAW, NIFTTL = makeNamespaces('NIFRAW', 'NIFTTL')

sgg = Graph(cache=True, verbose=True)
sgv = Vocabulary(cache=True)

# TODO future workflow: existing representation -> bags -> owl
# vastly preferred to the current nightmare
# the main things we need for the first arrow is a nice way to
# map to names of known phenotypes, so we need to load NNP and make sure
# for the second arrow we basically need to improve neuronManager

# consts
defined_class_parent = 'ilxtr:definedClassNeurons'
morpho_defined = 'ilxtr:definedClassNeuronsMorpho'
ephys_defined = 'ilxtr:definedClassNeuronsElectro'
morpho_phenotype =  'ilxtr:MorphologicalPhenotype'
morpho_edge = 'ilxtr:hasMorphologicalPhenotype'
ephys_phenotype = 'ilxtr:ElectrophysiologicalPhenotype'
ephys_edge = 'ilxtr:hasElectrophysiologicalPhenotype'
spiking_phenotype = 'ilxtr:SpikingPhenotype'
i_spiking_phenotype = 'ilxtr:PetillaInitialSpikingPhenotype'
s_spiking_phenotype = 'ilxtr:PetillaSustainedSpikingPhenotype'
spiking_edge = 'ilxtr:hasSpikingPhenotype'
fast_phenotype = 'ilxtr:FastSpikingPhenotype'
reg_phenotype = 'ilxtr:RegularSpikingNonPyramidalPhenotype'
expression_edge = 'ilxtr:hasExpressionPhenotype'
expression_defined = 'ilxtr:ExpressionClassifiedNeuron'
NIFCELL_NEURON = 'SAO:1417703748'  # 'NIFCELL:sao1417703748'

syntax = '{region}{layer_or_subregion}{expression}{ephys}{molecular}{morph}{cellOrNeuron}'
ilx_base = 'ILX:{:0>7}'

oq.OntCuries({'pheno': ilxtr['Phenotype/'],
              'MMRRC': 'http://www.mmrrc.org/catalog/getSDS.jsp?mmrrc_id='})
PREFIXES = {**makePrefixes('ilxtr',
                           'ILX',
                           'skos',
                           'owl',
                           'dc',
                           'nsu',
                           'CHEBI',
                           'NCBIGene',
                           'NCBITaxon',
                           'oboInOwl',
                           'NIFEXT',
                           'NIFRID',
                           'NLXCELL',
                           'SAO',
                           'UBERON',
                           'PR',),
            'TEMP':str(TEMP),
            'pheno':ilxtr['Phenotype/'],
}

def replace_object(find, replace, graph):  # note that this is not a sed 's/find/replace/g'
    find = graph.expand(find)
    for s, p, o in graph.g.triples((None, None, find)):
        graph.add_trip(s, p, replace)
        graph.g.remove((s, p, o))

def make_defined(graph, ilx_start, label, phenotype_id, restriction_edge, parent=None):
    ilx_start += 1
    id_ = TEMP['defined-neuron/' + label]
    id_ = graph.expand(id_)
    restriction = infixowl.Restriction(graph.expand(restriction_edge), graph=graph.g, someValuesFrom=phenotype_id)
    defined = infixowl.Class(id_, graph=graph.g)
    defined.label = rdflib.Literal(label + ' neuron')

    intersection = infixowl.BooleanClass(members=(graph.expand(NIFCELL_NEURON), restriction), graph=graph.g)
    #intersection = infixowl.BooleanClass(members=(restriction,), graph=self.graph.g)
    defined.equivalentClass = [intersection]
    if parent is not None:
        defined.subClassOf = [graph.expand(parent)]
    return ilx_start

    # TODO I think the defined class disjointness needs additional code... selecting from phenotype disjointness?
    #for dj in self.:
        #defined.disjointWith = [infixowl.Restriction(self.expand('ilxtr:hasPhenotype'), graph=self.graph.g, someValuesFrom=dj)]

def pprint(thing):
    for t in thing:
        print()
        for v in t:
            print(v)

def disjointUnionOf(graph, members):
    start = rdflib.BNode()
    current = start
    for member in members[:-1]:
        graph.add((current, rdflib.RDF.first, member))
        next_ = rdflib.BNode()
        graph.add((current, rdflib.RDF.rest, next_))
        current = next_

    graph.add((current, rdflib.RDF.first, members[-1]))
    graph.add((current, rdflib.RDF.rest, rdflib.RDF.nil))

    return start

def make_mutually_disjoint(graph, members):
    if len(members) > 1:
        first, rest = members[0], members[1:]
        for r in rest:
            log.debug(f'{first} {r}')
            graph.add_trip(first, rdflib.OWL.disjointWith, r)
        return make_mutually_disjoint(graph, rest)
    else:
        return members

def type_check(tup, types):
    return all([type(t) is ty for t, ty in zip(tup, types)])

def add_types(graph, phenotypes):  # TODO missing expression phenotypes! also basket type somehow :(
    """ Add disjoint union classes so that it is possible to see the invariants
        associated with individual phenotypes """

    collect = defaultdict(set)
    def recurse(id_, start, level=0):
        #print(level)
        for t in graph.g.triples((None, None, id_)):
            if level == 0:
                if t[1] != rdflib.term.URIRef('http://www.w3.org/2002/07/owl#someValuesFrom'):
                    continue
            if type_check(t, (rdflib.term.URIRef, rdflib.term.URIRef, rdflib.term.BNode)):
                #print(start, t[0])
                collect[start].add(t[0])
                return  #  we're done here, otherwise we hit instantiated subclasses
            if level > 1:
                if t[1] == rdflib.URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#first') or \
                   t[1] == rdflib.URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#rest'):
                    continue

            recurse(t[0], start, level + 1)


    for phenotype in phenotypes:
        recurse(phenotype, phenotype)

    return collect

def get_defined_classes(graph):
    phenotypes = [s for s, p, o in graph.g.triples((None, None, None)) if ' Phenotype' in o]
    inc = get_transitive_closure(graph, rdflib.RDFS.subClassOf, graph.expand('ilxtr:NeuronPhenotype'))
    collect = []
    def recurse(id_, start, level=0):
        if type(id_) == rdflib.term.URIRef and id_ != start:
            collect.append((start, id_))
            return
            #return start, id_
        for t in graph.g.triples((None, None, id_)):
            if level == 0 and type(t[0]) != rdflib.term.BNode:
                continue
            if level == 4:
                if t[1] !=  rdflib.term.URIRef('http://www.w3.org/2002/07/owl#equivalentClass'):
                    continue
            result = recurse(t[0], start, level + 1)
            #if result:  # this only works if there is a single equivalentClass axiom...
                #return result

    for id_ in inc:
        recurse(id_, id_)

    #for pheno, equiv in collect:
        #print(pheno, equiv, [_ for _ in graph.g.objects(equiv, rdflib.RDFS.label)])

    return collect

def get_transitive_closure(graph, edge, root):
    trips = set([t for t in graph.g.triples((None, edge, None))])
    valid_parents = {root}
    include = set()
    new_vp = set()
    old_len = -1
    output = set()
    output.update(valid_parents)
    while len(trips) != old_len:
        for t in trips:
            if t[2] in valid_parents:
                new_vp.add(t[0])
                include.add(t)
        valid_parents = new_vp
        output.update(valid_parents)
        old_len = len(trips)
        trips = trips - include

    return output


class OntId(OntId_):
    @property
    def u(self):
        return rdflib.URIRef(self)


def add_types(ng):
    graph = ng.g
    triples = (
        (ilxtr.NeuronCUT, rdf.type, owl.Class),
        (ilxtr.NeuronCUT, rdfs.subClassOf, OntId(_NEURON_CLASS).u),
        (ilxtr.NeuronEBM, rdf.type, owl.Class),
        (ilxtr.NeuronEBM, rdfs.subClassOf, OntId(_NEURON_CLASS).u),
    )
    [graph.add(t) for t in triples]


def phenotype_core_triples():
    yield ilxtr.delineates, owl.inverseOf, ilxtr.isDelineatedBy


def make_phenotypes():
    ilx_start = 50114

    olr = auth.get_path('ontology-local-repo')
    writeloc = olr / 'ttl'
    graph = makeGraph('phenotype-core',
                      writeloc=writeloc,
                      prefixes=PREFIXES)
    graph2 = makeGraph('phenotypes',
                       writeloc=writeloc,
                       prefixes=PREFIXES)

    defined_graph = createOntology(filename='NIF-Neuron-Defined',
                                   path='ttl/',
                                   prefixes=PREFIXES)

    edg = rdflib.Graph().parse(defined_graph.filename, format='turtle')
    defined_id_lookup = {o.value:s for s, o in edg.subject_objects(rdflib.RDFS.label)}
    log.debug(defined_id_lookup)

    # do edges first since we will need them for the phenotypes later
    # TODO real ilx_ids and use prefixes to manage human readability
    with open((resources / 'neuron_phenotype_edges.csv').as_posix(), 'rt') as f:
        rows = [r for r in csv.reader(f)]

    lookup = {
        'symmetric':'owl:SymmetricProperty',
        'asymmetric':'owl:AsymmetricProperty',
        'irreflexive':'owl:IrreflexiveProperty',
        'functional':'owl:FunctionalProperty',
    }
    pedges = set()
    for row in rows[1:]:
        if row[0].startswith('#') or not row[0]:
            if row[0] == '#references':
                break
            log.debug(row)
            continue
        id_ = ilxtr[row[0]]
        pedges.add(graph.expand('ilxtr:' + row[0]))
        graph.add_trip(id_, rdflib.RDFS.label, row[0])  # FIXME
        graph.add_trip(id_, rdflib.RDF.type, rdflib.OWL.ObjectProperty)
        if row[3]:
            graph.add_trip(id_, rdflib.namespace.SKOS.definition, row[3])
        if row[5]:
            graph.add_trip(id_, 'rdfs:comment', row[5])
        if row[6]:
            supers = row[6].split(',')
            for sup in supers:
                graph.add_trip(id_, rdflib.RDFS.subPropertyOf, 'ilxtr:' + sup)
        if row[7]:
            graph.add_trip(id_, rdflib.OWL.inverseOf, 'ilxtr:' + row[7])
        if row[8]:
            for t in row[8].split(','):
                t = t.strip()
                graph.add_trip(id_, rdflib.RDF.type, lookup[t])

    with open((resources / 'neuron_phenotype.csv').as_posix(), 'rt') as f:
        rows = [r for r in csv.reader(f) if any(r) and not r[0].startswith('#')]

    class PP(rowParse):  # FIXME use add_new in _row_post?
        SCD = 'subClassesDisjoint'
        DJW = 'disjointWith'
        def __init__(self):
            self.ilx_start = ilx_start
            self.parent_child_map = defaultdict(set)
            self.child_parent_map = defaultdict(set)
            self.scd = set()
            super().__init__(rows)

        def ilx_id(self, value):
            self.id_ = graph2.expand(value)
            self.Class = infixowl.Class(self.id_, graph=graph2.g)
            label = ' '.join(re.findall(r'[A-Z][a-z]*', self.id_.split(':')[1]))
            if 'Cone' in label or 'Rod' in label or 'Disc' in label:  # sigh hard coding
                label = label.replace(' Morphological', '')
            self._label = label.capitalize()

        def subClassOf(self, value):
            if value:
                values = value.split(',')
                for v in values:
                    self.parent = graph2.expand(v)
                    self.parent_child_map[self.parent].add(self.id_)
                    self.child_parent_map[self.id_].add(self.parent)
                    self.Class.subClassOf = [self.parent]

        def label(self, value):
            if value:
                self._label = value
                self.Class.label = rdflib.Literal(value)
            else:
                self.Class.label = rdflib.Literal(self._label)

        def notes(self, value):
            if value:
                graph2.add_trip(self.id_, 'rdfs:comment', value)

        def synonyms(self, value):
            if value:
                for v in value.split(','):
                    graph2.add_trip(self.id_, 'NIFRID:synonym', v)

        def rules(self, value):
            if value == PP.SCD:
                self.scd.add(self.id_)
            elif value.startswith(PP.DJW):
                [graph2.add_trip(self.id_, rdflib.OWL.disjointWith, _) for _ in value.split(' ')[1:]]

        def use_edge(self, value):
            if value:
                graph2.add_trip(self.id_, 'ilxtr:useObjectProperty', graph.expand('ilxtr:' + value))

        def _row_post(self):
            # defined class
            lookup = {
                graph.expand('ilxtr:AxonPhenotype'):rdflib.URIRef('http://axon.org'),
                graph.expand('ilxtr:AxonMorphologicalPhenotype'):None,
                graph.expand('ilxtr:DendritePhenotype'):rdflib.URIRef('http://dendrite.org'),
                graph.expand('ilxtr:DendriteMorphologicalPhenotype'):None,
                graph.expand('ilxtr:SomaPhenotype'):rdflib.URIRef('http://soma.org'),
                graph.expand('ilxtr:SomaMorphologicalPhenotype'):None,
                graph.expand('ilxtr:NeuronPhenotype'):graph.expand(NIFCELL_NEURON),  # we changed this structure...
                graph.expand('ilxtr:CellPhenotype'):None,
                #graph.expand('ilxtr:Phenotype'):graph.expand('ilxtr:Phenotype'),
                graph.expand('ilxtr:Phenotype'):graph.expand(NIFCELL_NEURON),  # FIXME ...
            }
            if self.id_ in lookup:
                return
            #elif 'Petilla' in self.id_:
                #return
            #else:
                #print(self.id_)

            # hidden label for consturctions
            hl = self._label.rsplit(' phenotype')[0]
            if hl.endswith('Morphological'):
                hl = hl.rsplit(' Morphological')[0]

            graph2.add_trip(self.id_, rdflib.namespace.SKOS.hiddenLabel, hl)

            label = rdflib.Literal(self._label.rstrip('phenotype') + 'neuron')

            self.ilx_start += 1
            #id_ = defined_graph.expand(ilx_base.format(self.ilx_start))
            # we have to look up the ids by label now
            if label.value in defined_id_lookup:
                id_ = defined_id_lookup[label.value]
            else:
                id_ = TEMP['newDefinedName' + str(self.ilx_start)]
            defined = infixowl.Class(id_, graph=defined_graph.g)
            defined.label = label
            #defined.label = rdflib.Literal(self._label.rstrip(' Phenotype') + ' neuron')  # the extra space in rstrip removes 'et ' as well WTF!
            #print(self._label)
            log.debug(f'_row_post ilx_start {self.ilx_start} {id_.rsplit("_")[-1]} {list(defined.label)[0]}')

            def getPhenotypeEdge(phenotype):
                log.debug(phenotype)
                edge = 'ilxtr:hasPhenotype'  # TODO in neuronManager...
                return edge
            edge = getPhenotypeEdge(self.id_)
            restriction = infixowl.Restriction(graph.expand(edge), graph=defined_graph.g, someValuesFrom=self.id_)

            parent = [p for p in self.child_parent_map[self.id_] if p]
            if parent:
                parent = parent[0]
                while 1:
                    if parent == defined_graph.expand('ilxtr:NeuronPhenotype'):
                        #defined.subClassOf = [graph.expand(defined_class_parent)]  # XXX this does not produce what we want
                        break
                    #else:
                        #print(parent, graph.expand('ilxtr:NeuronPhenotype'))

                    #print('xxxxxxxxxxxxxxxx', parent)
                    new_parent = [p for p in self.child_parent_map[parent] if p]
                    if new_parent:
                        parent = new_parent[0]
                    else:
                        break
                phenotype_equiv = lookup[parent]
            else:
                return


            intersection = infixowl.BooleanClass(members=(phenotype_equiv, restriction), graph=defined_graph.g)
            ##intersection = infixowl.BooleanClass(members=(restriction,), graph=self.graph.g)

            defined.equivalentClass = [intersection]

        def _end(self):
            for parent in self.scd:
                make_mutually_disjoint(graph2, list(self.parent_child_map[parent]))

    pp = PP()
    ilx_start = pp.ilx_start

    to_add = {}
    def lsn(word):
        rank = defaultdict(lambda:0)
        rank['PR'] = -100
        rank['NIFEXT'] = -50
        rank['UBERON'] = -10
        rank['NCBITaxon'] = -9
        #rank['NIFCELL'] = -8
        sort_rank = lambda r: rank[r['curie'].split(':')[0]]
        to_add[word] = graph2.expand(sorted(sgv.findByTerm(word), key=sort_rank)[0]['curie'])  # cheating

    # FIXME naming
    lsn('Parvalbumin')
    lsn('neuropeptide Y')
    lsn('VIP peptides')
    lsn('somatostatin')
    lsn('calbindin')
    lsn('calretinin')

    #for name, iri in to_add.items():  # XXX do not need, is already covered elsewhere
        #print('make_phenotypes ilx_start', ilx_start, name)
        #ilx_start = make_defined(defined_graph, ilx_start, name, iri, 'ilxtr:hasExpressionPhenotype', parent=expression_defined)

    #syn_mappings['calbindin'] = graph.expand('PR:000004967')  # cheating
    #syn_mappings['calretinin'] = graph.expand('PR:000004968')  # cheating
    #ontid = 'http://ontology.neuinfo.org/NIF/ttl/' + graph.name + '.ttl'
    ontid = NIFRAW['neurons/ttl/' + graph.name + '.ttl']
    graph.add_ont(ontid, 'NIF Phenotype core', comment= 'This is the core set of predicates used to model phenotypes and the parent class for phenotypes.')
    graph.add_class('ilxtr:Phenotype', label='Phenotype')
    graph.add_trip('ilxtr:Phenotype', 'skos:definition', 'A Phenotype is a binary property of a biological entity. Phenotypes are derived from measurements made on the subject of interest. While Phenotype is not currently placed within the BFO hierarchy, if we were to place it, it would fall under BFO:0000016 -> disposition, since these phenotypes are contingent on the experimental conditions under which measurements were made and are NOT qualities. For consideration: in theory this would mean that disjointness does not make sense, even for things that would seem to be obviously disjoint such as Accommodating and Non-Accommodating. However, this information can still be captured on a subject by subject basis by asserting that for this particular entity, coocurrance of phenotypes is not possible. This still leaves the question of whether the class of biological entities that correspond to the bag of phenotypes is implicitly bounded/limited only to the extrinsic and unspecified experimental conditions, some of which are not and cannot be included in a bag of phenotypes. The way to deal with this when we want to include 2 \'same time\' disjoint phenotypes, is to use a logical phenotype to wrap them with an auxiliary variable that we think accounts for the difference.')
    #graph.add_trip(ontid, rdflib.RDFS.comment, 'The NIF Neuron ontology holds materialized neurons that are collections of phenotypes.')
    #graph.add_trip(ontid, rdflib.OWL.versionInfo, ONTOLOGY_DEF['version'])
    #graph.g.commit()
    #get_defined_classes(graph)  # oops...
    add_types(graph)
    [graph.g.add(t) for t in phenotype_core_triples()]
    graph.write()  # moved below to incorporate uwotm8

    #ontid2 = 'http://ontology.neuinfo.org/NIF/ttl/' + graph2.name + '.ttl'
    ontid2 = NIFRAW['neurons/ttl/' + graph2.name + '.ttl']
    graph2.add_ont(ontid2, 'NIF Phenotypes', comment='A taxonomy of phenotypes used to model biological types as collections of measurements.')
    graph2.add_trip(ontid2, 'owl:imports', ontid)
    graph2.write()

    syn_mappings = {}
    for sub, syn in ([_ for _ in graph2.g.subject_objects(graph2.expand('NIFRID:synonym'))] +
                     [_ for _ in graph2.g.subject_objects(rdflib.RDFS.label)]):
        syn = syn.toPython()
        if syn in syn_mappings:
            log.error(f'duplicate synonym! {syn} {sub}')
        syn_mappings[syn] = sub

    phenotypes_o = [(s, o) for s, p, o in graph2.g.triples((None, None, None)) if ' Phenotype' in o]
    extras = {o.replace('Phenotype', '').strip():s
              for s, o in phenotypes_o}
    epl = {(o.replace('Petilla', '')
            .replace('Initial', '')
            .replace('Sustained', '')
            .replace('Spiking', '')
            .strip()):s for o, s in extras.items()}
    extras.update(epl)
    el = {o.lower():s for o, s in extras.items()}
    extras.update(el)
    syn_mappings.update(extras)
    phenotypes = [s for s, o in phenotypes_o]
    inc = get_transitive_closure(graph2, rdflib.RDFS.subClassOf, graph2.expand('ilxtr:NeuronPhenotype'))  # FIXME not very configurable...

    return syn_mappings, pedges, ilx_start, inc, defined_graph

def _rest_make_phenotypes():
    glb = auth.get_path('git-local-base')
    olr = auth.get_path('ontology-local-repo')
    #phenotype sources
    neuroner = (glb /
                'neuroNER/resources/bluima/neuroner/hbp_morphology_ontology.obo').as_posix()
    neuroner1 = (glb /
                 'neuroNER/resources/bluima/neuroner/hbp_electrophysiology_ontology.obo').as_posix()
    neuroner2 = (glb /
                 'neuroNER/resources/bluima/neuroner/hbp_electrophysiology-triggers_ontology.obo').as_posix()
    nif_qual = (olr / 'ttl/NIF-Quality.ttl').as_posix()

    mo = OboFile(os.path.expanduser(neuroner))
    mo1 = OboFile(os.path.expanduser(neuroner1))
    mo2 = OboFile(os.path.expanduser(neuroner2))
    mo_ttl = mo.__ttl__() + mo1.__ttl__() + mo2.__ttl__()

    mo_ttl = """\
    @prefix : <http://FIXME.org/> .
    @prefix nsu: <http://www.FIXME.org/nsupper#> .
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
    @prefix owl: <http://www.w3.org/2002/07/owl#> .
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
    """ + mo_ttl

    #sio = io.StringIO()
    #sio.write(mo_ttl)

    ng = rdflib.Graph()
    ng.parse(data=mo_ttl, format='turtle')
    ng.parse(os.path.expanduser(nif_qual), format='turtle')
    #ng.namespace_manager.bind('default1', None, override=False, replace=True)
    ng.remove((None, rdflib.OWL.imports, None))

    bad_match = {
        'http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#nlx_qual_20090505',
        'http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#sao1693353776',
        'http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#sao1288413465',
        'http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#sao4459136323',
        'http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#nlx_qual_20090507',
    }

    exact = []
    similar = []
    quals = []
    s2 = {}

    for subject, label in sorted(ng.subject_objects(rdflib.RDFS.label)):

        syns = set([a for a in ng.objects(subject, rdflib.URIRef('http://www.FIXME.org/nsupper#synonym'))])
        syns.update(set([a for a in ng.objects(subject, rdflib.URIRef('http://ontology.neuinfo.org/NIF/Backend/OBO_annotation_properties.owl#synonym'))]))
        #if syns:
            #print(syns)
        #print(subject)
        #print(label.lower())
        if 'quality' in label.lower():
            quals.append((subject, label))
        subpre = ng.namespace_manager.compute_qname(subject)[1]
        llower = rdflib.Literal(label.lower(), lang='en')
        for s in ng.subjects(rdflib.RDFS.label, llower):
            if s != subject:
                exact.append((subject, s, label, llower))
        for s, p, o in sorted(ng.triples((None, rdflib.RDFS.label, None))):
            spre = ng.namespace_manager.compute_qname(s)[1]
            if subject != s and label.lower() in o.lower().split(' ') and spre != subpre:
                if s.toPython() in bad_match or subject.toPython() in bad_match:
                    continue
                #print()
                #print(spre, subpre)
                similar.append((subject, s, label, o))
                if subpre.toPython() == 'http://FIXME.org/':
                    print('YAY')
                    print(label, ',', o)
                    print(subject, s)
                    subject, s = s, subject
                    label, o = o, label

                if subject in s2:
                    #print('YES IT EXISTS')
                    #print(syns, label, [subject, s])
                    s2[subject]['syns'].update(syns)
                    s2[subject]['syns'].add(label)
                    s2[subject]['xrefs'] += [subject, s]
                else:
                    s2[subject] = {'label': label.toPython(), 'o': o.toPython(), 'xrefs':[subject, s], 'syns':syns}  # FIXME overwrites

    pprint(quals)
    """ print stuff
    print('matches')
    pprint(exact)
    pprint(similar)

    #print('EXACT', exact)

    print()
    for k, v in s2.items():
        print(k)
        for k, v2 in sorted(v.items()):
            print('    ', k, ':', v2)
    #"""

    desired_nif_terms = set() #{
        #'NIFQUAL:sao1959705051',  # dendrite
        #'NIFQUAL:sao2088691397',  # axon
        #'NIFQUAL:sao1057800815',  # morphological
        #'NIFQUAL:sao-1126011106',  # soma
        #'NIFQUAL:',
        #'NIFQUAL:',
    #}
    starts = [
        #"NIFQUAL:sao2088691397",
        #"NIFQUAL:sao1278200674",
        #"NIFQUAL:sao2088691397",
        #"NIFQUAL:sao-1126011106",  # FIXME WTF IS THIS NONSENSE  (scigraph bug?)
        quote("http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#sao1959705051").replace('/','%2F'),
        quote("http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#sao2088691397").replace('/','%2F'),
        quote("http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#sao1278200674").replace('/','%2F'),
        quote("http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#sao2088691397").replace('/','%2F'),
        quote("http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#sao-1126011106").replace('/','%2F'),
    ]

    for id_ in starts:
        want = sgg.getNeighbors(id_, relationshipType='subClassOf', direction='INCOMING', depth=5)
        #print(id_, want)
        desired_nif_terms.update([n['id'] for n in want['nodes']])

    print(desired_nif_terms)


    ilx_start = 50114
    print(ilx_base.format(ilx_start))
    new_terms = {}
    dg = makeGraph('uwotm8', prefixes=PREFIXES)
    xr = makeGraph('xrefs', prefixes=PREFIXES)
    for s, o in sorted(ng.subject_objects(rdflib.RDFS.label))[::-1]:
        spre = ng.namespace_manager.compute_qname(s)[1]
        #if spre.toPython() == g.namespaces['NIFQUAL']:
            #print('skipping', s)
            #continue  # TODO
        if s in new_terms:
            print(s, 'already in as xref probably')
            continue
        #elif spre.toPython() != 'http://uri.interlex.org/base/ilx_' or spre.toPython() != 'http://FIXME.org/' and s.toPython() not in desired_nif_terms:
        #elif spre.toPython() != 'http://FIXME.org/' and s.toPython() not in desired_nif_terms:
            #print('DO NOT WANT', s, spre)
            #continue

        syns = set([s for s in ng.objects(s, dg.namespaces['nsu']['synonym'])])
        #data['syns'] += syns

        data = {}
        id_ = ilx_base.format(ilx_start)
        ilx_start += 1
        if s in s2:
            d = s2[s]
            syns.update(d['syns'])
            new_terms[d['xrefs'][0]] = {'replaced_by':id_}
            xr.add_trip(d['xrefs'][0], 'oboInOwl:replacedBy', id_)
            #dg.add_trip(d['xrefs'][0], 'oboInOwl:replacedBy', id_)
            new_terms[d['xrefs'][1]] = {'replaced_by':id_}
            xr.add_trip(d['xrefs'][1], 'oboInOwl:replacedBy', id_)
            #dg.add_trip(d['xrefs'][1], 'oboInOwl:replacedBy', id_)

            data['labels'] = [d['label'], d['o']]
            #dg.add_trip(id_, rdflib.RDFS.label, d['label'])
            dg.add_trip(id_, rdflib.RDFS.label, d['o'])
            data['xrefs'] = d['xrefs']
            for x in d['xrefs']:  # FIXME... expecting order of evaluation errors here...
                dg.add_trip(id_, 'oboInOwl:hasDbXref', x)  # xr
                xr.add_trip(id_, 'oboInOwl:hasDbXref', x)  # x

        elif (spre.toPython() != 'http://ontology.neuinfo.org/NIF/BiomaterialEntities/NIF-Quality.owl#' or
              ng.namespace_manager.qname(s).replace('default1','NIFQUAL') in desired_nif_terms):  # skip non-xref quals
            #print(ng.namespace_manager.qname(s).replace('default1','NIFQUAL'))
            new_terms[s] = {'replaced_by':id_}
            xr.add_trip(s, 'oboInOwl:replacedBy', id_)
            data['labels'] = [o.toPython()]
            dg.add_trip(id_, rdflib.RDFS.label, o.toPython())
            data['xrefs'] = [s]
            dg.add_trip(id_, 'oboInOwl:hasDbXref', s)  # xr
            xr.add_trip(id_, 'oboInOwl:hasDbXref', s)  # xr
        else:
            ilx_start -= 1
            continue

        new_terms[id_] = data
        dg.add_trip(id_, rdflib.RDF.type, rdflib.OWL.Class)
        xr.add_trip(id_, rdflib.RDF.type, rdflib.OWL.Class)
        for syn in syns:
            if syn.toPython() not in data['labels']:
                if len(syn) > 3:
                    dg.add_trip(id_, 'NIFRID:synonym', syn)
                elif syn:
                    dg.add_trip(id_, 'NIFRID:abbrev', syn)

        if 'EPHYS' in s or any(['EPHYS' in x for x in data['xrefs']]):
            dg.add_trip(id_, rdflib.RDFS.subClassOf, ephys_phenotype)
        elif 'MORPHOLOGY' in s or any(['MORPHOLOGY' in x for x in data['xrefs']]):
            dg.add_trip(id_, rdflib.RDFS.subClassOf, morpho_phenotype)

    #dg.write(convert=False)
    xr.write(convert=False)

    #skip this for now, we can use DG to do lookups later
    #for t in dg.g.triples((None, None, None)):
        #g.add_trip(*t)  # only way to clean prefixes :/
    add_phenotypes(g)
    g.write(convert=False)

    g2 = makeGraph('pheno-comp', PREFIXES)
    for t in ng.triples((None, None, None)):
        g2.add_trip(*t)  # only way to clean prefixes :/

    g2.write(convert=False)

    syn_mappings = {}
    for sub, syn in [_ for _ in g.g.subject_objects(g.expand('NIFRID:synonym'))] + [_ for _ in g.g.subject_objects(rdflib.RDFS.label)]:
        syn = syn.toPython()
        if syn in syn_mappings:
            log.error(f'duplicate synonym! {syn} {sub}')
        syn_mappings[syn] = sub

    #breakpoint()
    return syn_mappings, pedges, ilx_start

def make_neurons(syn_mappings, pedges, ilx_start_, defined_graph, old=False):
    ilx_start = ilx_start_
    cheating = {'vasoactive intestinal peptide':'VIP',
                'star':None,  # is a morphological phen that is missing but hits scigraph
               }
    ng = createOntology(filename='NIF-Neuron',
                        path='ttl/',
                        prefixes=PREFIXES)

    #""" It seemed like a good idea at the time...
    olr = auth.get_path('ontology-local-repo')
    nif_cell = (olr / 'ttl/NIF-Cell.ttl').as_posix()  # need to be on neurons branch
    cg = rdflib.Graph()
    cg.parse(os.path.expanduser(nif_cell), format='turtle')
    missing = (
        'NIFEXT:55',
        'NIFEXT:56',
        'NIFEXT:57',
        'NIFEXT:59',
        'NIFEXT:81',
        'NLXCELL:091205',
        NIFCELL_NEURON,
        'SAO:2128417084',
        'SAO:862606388',  # secondary, not explicitly in the hbp import
    )
    for m in missing:
        m = ng.expand(m)
        for s, p, o in cg.triples((m, None, None)):
            ng.add_trip(s, p, o)



    #cg.remove((None, rdflib.OWL.imports, None))  # DONOTWANT NIF-Cell imports
    #for t in cg.triples((None, None, None)):
        #ng.add_trip(*t)  # only way to clean prefixes :/
    #cg = None
    #"""

    hbp_cell = (olr / 'ttl/generated/NIF-Neuron-HBP-cell-import.ttl').as_posix()  # need to be on neurons branch
    _temp = rdflib.Graph()  # use a temp to strip nasty namespaces
    _temp.parse(os.path.expanduser(hbp_cell), format='turtle')
    for s, p, o in _temp.triples((None,None,None)):
        if s != rdflib.URIRef('http://ontology.neuinfo.org/NIF/ttl/generated/NIF-Neuron-HBP-cell-import.ttl'):
            ng.g.add((s,p,o))

    base = 'http://ontology.neuinfo.org/NIF/ttl/'

    ontid = base + ng.name + '.ttl'
    ng.add_trip(ontid, rdflib.RDF.type, rdflib.OWL.Ontology)
    ng.add_trip(ontid, rdflib.OWL.imports, base + 'phenotypes.ttl')
    ng.add_trip(ontid, rdflib.OWL.imports, base + 'NIF-Neuron-Defined.ttl')
    ng.add_trip(ontid, rdflib.OWL.imports, base + 'hbp-special.ttl')
    #ng.add_trip(ontid, rdflib.OWL.imports, base + 'NIF-Cell.ttl')  # NO!
    #ng.add_trip(ontid, rdflib.OWL.imports, base + 'external/uberon.owl')
    #ng.add_trip(ontid, rdflib.OWL.imports, base + 'external/pr.owl')
    ng.replace_uriref('ilxtr:hasMolecularPhenotype', 'ilxtr:hasExpressionPhenotype')

    #defined_graph = makeGraph('NIF-Neuron-Defined', prefixes=PREFIXES, graph=_g)
    #defined_graph.add_trip(base + defined_graph.name + '.ttl', rdflib.RDF.type, rdflib.OWL.Ontology)
    defined_graph.add_trip(defined_graph.ontid, rdflib.OWL.imports, base + 'phenotypes.ttl')

    #log.debug(str(list(syn_mappings)))
    done = True#False
    done_ = set()
    for pedge in pedges:
        for s, p, o_lit in ng.g.triples((None, pedge, None)):
            o = o_lit.toPython()
            success = False
            true_o = None
            true_id = None
            terms = []
            _pp = p.toPython()
            if o in syn_mappings:
                id_ = syn_mappings[o]  # FIXME can this happen more than once?

                ng.add_hierarchy(id_, p, s)
                ng.g.remove((s, p, o_lit))
                #print('SUCCESS, substituting', o, 'for', id_)
                success = True
                true_o = o_lit
                true_id = id_

            elif ('Location' in _pp or 'LocatedIn' in _pp or 'ElementsIn' in
                  _pp or 'TerminalsIn' in _pp):  # lift location to restrictions
                if o.startswith('http://'):
                    ng.add_hierarchy(o_lit, p, s)
                    ng.g.remove((s, p, o_lit))

                    data = sgv.findById(o)
                    label = data['labels'][0]
                    ng.add_trip(o, rdflib.RDF.type, rdflib.OWL.Class)
                    ng.add_trip(o, rdflib.RDFS.label, label)

                    success = True
                    true_o = label
                    true_id = o_lit

            else:
                if o in cheating:
                    o = cheating[o]
                    if o is None:
                        log.debug(f'{o_lit}')
                        continue

                terms = [t for t in OntTerm.query(term=o)]
                for t in terms:
                    if t.prefix in ('PR', 'CHEBI'):
                        sgt = t.URIRef
                        ng.add_hierarchy(sgt, p, s)
                        ng.g.remove((s, p, o_lit))

                        label = t.label
                        ng.add_trip(sgt, rdflib.RDF.type, rdflib.OWL.Class)
                        ng.add_trip(sgt, rdflib.RDFS.label, label)

                        success = True
                        true_o = label
                        true_id = sgt
                        break

                if not success:
                    for t in terms:
                        if t.prefix == 'NIFEXT':
                            sgt = t.URIRef
                            ng.add_hierarchy(sgt, p, s)
                            ng.g.remove((s, p, o_lit))

                            label = t.label
                            ng.add_trip(sgt, rdflib.RDF.type, rdflib.OWL.Class)
                            ng.add_trip(sgt, rdflib.RDFS.label, label)

                            success = True
                            true_o = label
                            true_id = sgt
                            break

                    else:
                        if terms:
                            log.warning(f'Did not succeed for {o!r} found\n{terms}')

            if o not in done_ and success:
                done_.add(o)
                t = tuple(defined_graph.g.triples((None, rdflib.OWL.someValuesFrom, true_id)))
                if t:
                    # these are 1:1 defined neurons from phenotypes so should only show up once
                    #log.debug(f'ALREADY IN {t}')
                    pass
                else:
                    ilx_start += 1
                    id_ = TEMP[str(ilx_start)]
                    defined_graph.add_trip(id_, rdflib.RDF.type, rdflib.OWL.Class)
                    restriction = infixowl.Restriction(p, graph=defined_graph.g, someValuesFrom=true_id)
                    intersection = infixowl.BooleanClass(members=(defined_graph.expand(NIFCELL_NEURON), restriction), graph=defined_graph.g)
                    this = infixowl.Class(id_, graph=defined_graph.g)
                    this.equivalentClass = [intersection]
                    this.subClassOf = [defined_graph.expand(defined_class_parent)]
                    this.label = rdflib.Literal(true_o + ' neuron')
                    log.info(f'make_neurons ilx_start {ilx_start} {list(this.label)[0]}')
                    if not done:
                        breakpoint()
                        done = True

            elif not success:
                log.warning(f'failures for {o} -> {terms}')

    defined_graph.add_class(defined_class_parent, NIFCELL_NEURON, label='defined class neuron')
    defined_graph.add_trip(defined_class_parent, rdflib.namespace.SKOS.definition, 'Parent class For all defined class neurons')

    if old:
        defined_graph.write()  # NIF-Neuron-Defined
        ng.write()  # NIF-Neuron

    for sub, syn in [_ for _ in ng.g.subject_objects(ng.expand('NIFRID:synonym'))] + [_ for _ in ng.g.subject_objects(rdflib.RDFS.label)]:
        syn = syn.toPython()
        if syn in syn_mappings:
            log.error(f'duplicate synonym! {syn} {sub}')
        syn_mappings[syn] = sub

    return ilx_start

def add_phenotypes(graph):

    cell_phenotype = 'ilxtr:CellPhenotype'
    neuron_phenotype = 'ilxtr:NeuronPhenotype'
    #ephys_phenotype = 'ilxtr:ElectrophysiologicalPhenotype'
    #spiking_phenotype = 'ilxtr:SpikingPhenotype'
    #i_spiking_phenotype = 'ilxtr:PetillaInitialSpikingPhenotype'
    burst_p = 'ilxtr:PetillaInitialBurstSpikingPhenotype'
    classical_p = 'ilxtr:PetillaInitialClassicalSpikingPhenotype'
    delayed_p = 'ilxtr:PetillaInitialDelayedSpikingPhenotype'
    #s_spiking_phenotype = 'ilxtr:PetillaSustainedSpikingPhenotype'
    #morpho_phenotype = 'ilxtr:MorphologicalPhenotype'
    ac_p = 'ilxtr:PetillaSustainedAccommodatingPhenotype'
    nac_p = 'ilxtr:PetillaSustainedNonAccommodatingPhenotype'
    st_p = 'ilxtr:PetillaSustainedStutteringPhenotype'
    ir_p = 'ilxtr:PetillaSustainedIrregularPhenotype'

    #fast = 'ilxtr:FastSpikingPhenotype'
    #reg_int = 'ilxtr:RegularSpikingNonPyramidalPhenotype'

    graph.add_class(cell_phenotype, autogen=True)
    graph.add_class(neuron_phenotype, cell_phenotype, autogen=True)
    graph.add_class(ephys_phenotype, neuron_phenotype, autogen=True)
    graph.add_class(spiking_phenotype, ephys_phenotype, autogen=True)
    graph.add_class(fast_phenotype, spiking_phenotype,('Fast spiking'), autogen=True)
    graph.add_class(reg_phenotype, spiking_phenotype,('Non-fast spiking','Regular spiking non-pyramidal'), autogen=True)
    graph.add_class(i_spiking_phenotype, spiking_phenotype, autogen=True)
    iClass = infixowl.Class(graph.expand(i_spiking_phenotype), graph=graph.g)

    graph.add_class(burst_p, i_spiking_phenotype, ('burst',), autogen=True)
    graph.add_class(classical_p, i_spiking_phenotype, ('classical',), autogen=True)
    graph.add_class(delayed_p, i_spiking_phenotype, ('delayed',), autogen=True)
    graph.add_class(s_spiking_phenotype, spiking_phenotype, autogen=True)
    sClass = infixowl.Class(graph.expand(s_spiking_phenotype), graph=graph.g)
    sClass.disjointWith = [iClass]

    graph.add_class(ac_p, s_spiking_phenotype, ('accommodating',), autogen=True)  # FIXME this is silly
    graph.add_class(nac_p, s_spiking_phenotype, ('non accommodating',), autogen=True)
    graph.add_class(st_p, s_spiking_phenotype, ('stuttering',), autogen=True)
    graph.add_class(ir_p, s_spiking_phenotype, ('irregular',), autogen=True)
    graph.add_class(morpho_phenotype, neuron_phenotype, autogen=True)

class table1:
    pass # dead use neuron_ma2015.py

def make_table1(syn_mappings, ilx_start, phenotypes):
    # TODO when to explicitly subClassOf? I think we want this when the higher level phenotype bag is shared
    # it may turn out that things like the disjointness exist at a higher level while correlated properties
    # should be instantiated together as sub classes, for example if cck and
    # FIXME disagreement about nest basket cells
    # TODO hasPhenotypes needs to be function to get phenotypeOf to work via reasoner??? this seems wrong.
    #  this also works if phenotypeOf is inverseFunctional
    #  hasPhenotype shall be asymmetric, irreflexive, and intransitive
    # XXX in answer to Maryann's question about why we need the morphological phenotypes by themselves:
    #  if we don't have them we can't agregate across orthogonal phenotypes since owl correctly keeps the classes distinct
    # TODO disjointness axioms work really well on defined classes and propagate excellently
    # TODO add 'Petilla' or something like that to the phenotype definitions
    #  we want this because 'Petilla' denotes the exact ANALYSIS used to determine the phenotype
    #  there are some additional 'protocol' related restrictions on what you can apply analysis to
    #  but we don't have to model those explicitly which would be a nightmare and break the
    #  orthogonality of the cell type decomposition
    # TODO to make this explicit we need to include that phenotypes require 2 things
    #  1) a type of data (data type?) 2) a way to classify that data (analysis protocol)
    #
    # need a full type restriction... property chain?

    graph = makeGraph('hbp-special', prefixes=PREFIXES)  # XXX fix all prefixes

    with open((resources / '26451489 table 1.csv').as_posix(), 'rt') as f:
        rows = [list(r) for r in zip(*csv.reader(f))]

    base = 'http://ontology.neuinfo.org/NIF/ttl/'
    ontid = base + graph.name + '.ttl'
    graph.add_trip(ontid, rdflib.RDF.type, rdflib.OWL.Ontology)
    graph.add_trip(ontid, rdflib.OWL.imports, base + 'phenotypes.ttl')
    graph.add_trip(ontid, rdflib.OWL.imports, base + 'NIF-Neuron-Defined.ttl')

    def lsn(word):
        syn_mappings[word] = graph.expand(sgv.findByTerm(word)[0]['curie'])  # cheating
    lsn('Parvalbumin')
    lsn('neuropeptide Y')
    lsn('VIP peptides')
    lsn('somatostatin')
    syn_mappings['calbindin'] = graph.expand('PR:000004967')  # cheating
    syn_mappings['calretinin'] = graph.expand('PR:000004968')  # cheating
    t = table1(graph, rows, syn_mappings, ilx_start)
    ilx_start = t.ilx_start

    # adding fake mouse data
    #with open(refile(__file__, 'resources/26451489 table 1.csv'), 'rt') as f:  # FIXME annoying
        #rows = [list(r) for r in zip(*csv.reader(f))]
    #t2 = table1(graph, rows, syn_mappings, ilx_start, species='NCBITaxon:10090')  # FIXME double SOM+ phenos etc
    #ilx_start = t2.ilx_start

    def do_graph(d):
        sgt = graph.expand(d['curie'])
        label = d['labels'][0]
        graph.add_trip(sgt, rdflib.RDF.type, rdflib.OWL.Class)
        graph.add_trip(sgt, rdflib.RDFS.label, label)

    done = set()
    for s, p, o in graph.g.triples((None, None, None)): #(rdflib.RDFS.subClassOf,rdflib.OWL.Thing)):
        if o not in done and type(o) == rdflib.term.URIRef:
            done.add(o)
            if not [_ for _ in graph.g.objects((o, rdflib.RDFS.label))]:
                d = sgv.findById(o)
                if d:
                    if 'PR:' in d['curie']:
                        do_graph(d)
                    elif 'NIFMOL:' in d['curie']:
                        do_graph(d)
                    elif 'UBERON:' in d['curie']:
                        do_graph(d)
                    elif 'NCBITaxon:' in d['curie']:
                        do_graph(d)
                    elif 'NIFCELL:' in d['curie']:
                        do_graph(d)


    # FIXME this is a dupe with defined_class
    #graph.add_trip(defined_class_parent, rdflib.RDF.type, rdflib.OWL.Class)
    #graph.add_trip(defined_class_parent, rdflib.RDFS.label, 'defined class neuron')
    #graph.add_trip(defined_class_parent, rdflib.namespace.SKOS.description, 'Parent class For all defined class neurons')
    #graph.add_trip(defined_class_parent, rdflib.RDFS.subClassOf, NIFCELL_NEURON)
    #graph.add_trip(morpho_defined, rdflib.RDFS.subClassOf, defined_class_parent)
    #graph.add_trip(morpho_defined, rdflib.RDFS.label, 'Morphologically classified neuron')  # FIXME -- need asserted in here...
    #graph.add_trip(ephys_defined, rdflib.RDFS.subClassOf, defined_class_parent)
    #graph.add_trip(ephys_defined, rdflib.RDFS.label, 'Electrophysiologically classified neuron')

    graph.add_class(expression_defined, NIFCELL_NEURON, autogen=True)
    graph.add_class('ilxtr:NeuroTypeClass', NIFCELL_NEURON, label='Neuron TypeClass')

    graph.g.commit()

    phenotype_dju_dict = add_types(graph, phenotypes)
    for pheno, disjoints in phenotype_dju_dict.items():
        name = ' '.join(re.findall(r'[A-Z][a-z]*', pheno.split(':')[1])[:-1])  #-1: drops Phenotype
        ilx_start += 1# = make_defined(graph, ilx_start, name + ' neuron type', pheno, 'ilxtr:hasPhenotype')
        id_ = TEMP[str(ilx_start)]
        typeclass = infixowl.Class(id_, graph=graph.g)
        typeclass.label = rdflib.Literal(name + ' neuron type')

        restriction = infixowl.Restriction(graph.expand('ilxtr:hasPhenotype'), graph=graph.g, someValuesFrom=pheno)
        #typeclass.subClassOf = [restriction, graph.expand('ilxtr:NeuroTypeClass')]

        ntc = graph.expand('ilxtr:NeuroTypeClass')
        intersection = infixowl.BooleanClass(members=(ntc, restriction), graph=graph.g)
        typeclass.equivalentClass = [intersection]

        # FIXME not clear that we should be doing typeclasses this way.... :/
        # requires more thought, on the plus side you do get better reasoning...
        disjointunion = disjointUnionOf(graph=graph.g, members=list(disjoints))
        graph.add_trip(id_, rdflib.OWL.disjointUnionOf, disjointunion)


    graph.write()
    return t
    #print(t._set_Electrical_types)
    #_ = [[print(v) for v in [k] + list(v) + ['\n']] for k,v in t.__dict__.items() if '_set_' in k]

#def pattern_match(start, chain):   # sparql?
    # start -> predicate1 -> anon.Class -> predicate2 -> ( -> anon.Restriction -> predicate3 -> match
    # match <- predicate3 <- [predicate2 <- ([predicate1 <- start

def expand_syns(syn_mappings):
    for syn, iri in list(syn_mappings.items()):
        lower = syn.lower()
        lower_less_phenotype = lower.rsplit(' phenotype')[0]
        syn_mappings[lower] = iri
        #print()
        #print(lower)
        #print(lower_less_phenotype)
        #print()
        if lower_less_phenotype != lower:
            syn_mappings[lower_less_phenotype] = iri

def predicate_disambig(graph):
    # the useful idiot holds 'appriopriate' predicate/phenotype mappings where there may be some ambiguity (e.g. brain region vs layer) it is OK to hardcode these
    # the ui is not an owl:Class
    ui = ilxtr.PhenotypePredicateDisambiguation
    def uit(p, o):
        out = (ui, graph.expand(p), graph.expand(o))
        graph.add_trip(*out)
        return out

    uit('ilxtr:hasLayerLocation', 'UBERON:0005390')
    uit('ilxtr:hasLayerLocation', 'UBERON:0005391')
    uit('ilxtr:hasLayerLocation', 'UBERON:0005392')
    uit('ilxtr:hasLayerLocation', 'UBERON:0005393')
    uit('ilxtr:hasLayerLocation', 'UBERON:0005394')
    uit('ilxtr:hasLayerLocation', 'UBERON:0005395')


def make_models():
    from importlib import import_module
    from neurondm.models import __all__
    # skip pcl for now since it is unlifted
    skip = 'allen_type_specimen_mapping', 'phenotype_direct', 'pcl',
    skip += 'basic_neurons', # 'allen_cell_types',
    __all__ = [a for a in __all__ if a not in skip]
    for module in __all__:
        if 'CI' in os.environ and module == 'cuts':  # FIXME XXX temp fix
            continue
        m = import_module(f'neurondm.models.{module}')
        #m = import_module(f'neurondm.compiled.{module}')  # XXX
        if not hasattr(m, 'config') and hasattr(m, 'main'):
            config, *_ = m.main()  # FIXME cuts and allen ct hack
        else:
            config = m.config


def make_bridge():
    from pyontutils.utils import subclasses
    from pyontutils.core import Ont, build
    from neurondm.lang import Config, NeuronEBM  # need ebm for subclasses to work
    from neurondm.models import __all__
    #from neurondm.compiled import __all__  # XXX
    skip = 'phenotype_direct',
    __all__ = [a for a in __all__ if a not in skip]
    log.info('building the following\n' + '\n'.join(__all__))
    models_imports = 'allen', 'markram', 'huang', 'common', 'Defined', 'bolser',
    #models_imports = 'markram', 'huang', 'common', 'Defined'

    class neuronBridge(Ont):
        """ Main bridge for importing the various files that
            make up the neuron phenotype ontology. """

        remote_base = str(NIFRAW['neurons/'])
        path = 'ttl/bridge/'
        filename = 'neuron-bridge'
        name = 'Neuron Bridge'
        imports = chain((subclass.iri for subclass in subclasses(Config)
                         if anyMembers(subclass.iri, *models_imports)),
                        (NIFRAW['neurons/ttl/generated/uberon-parcellation-mappings.ttl'],
                         NIFRAW['neurons/ttl/generated/neurons/cut-roundtrip.ttl'],
                        ))

    nb, *_ = build(neuronBridge, n_jobs=1)
    #log.critical(f'{nb.imports} {models_imports}')

def make_devel():
    from ttlser import CustomTurtleSerializer
    from pyontutils import combinators as cmb
    from pyontutils.core import Ont, build
    from pyontutils.utils import Async, deferred

    # inefficient but thorough way to populate the subset of objects we need.
    terms = set()

    olr = auth.get_path('ontology-local-repo')
    n = (olr / 'ttl/generated/neurons')
    fns =('allen-cell-types.ttl',
          'cut-development.ttl',
          'common-usage-types.ttl',
          #'cut-roundtrip.ttl',
          'huang-2017.ttl',
          'markram-2015.ttl',

          #'bolser-lewis.ttl',
          'keast-2020.ttl',
          'apinatomy-neuron-populations.ttl',
          #'apinatomy-locations.ttl',
          #'nerves.ttl',
          'sparc-nlp.ttl',
          'apinat-simple-sheet.ttl',
          )
    g = OntConjunctiveGraph()
    for fn in fns:
        fp = n / fn
        g.parse(fp.as_posix(), format='ttl')

    _pi = (olr / 'ttl/phenotype-indicators.ttl')
    g.parse(_pi.as_posix(), format='ttl')

    bads = ('TEMP', 'TEMPIND', 'ilxtr', 'rdf', 'rdfs', 'owl', '_', 'prov', 'BFO1SNAP', 'NLXANAT',
            'NIFRAW', 'NLXCELL', 'NLXNEURNT', 'BFO', 'MBA', 'JAX', 'MMRRC', 'ilx', 'CARO', 'NLX',
            'BIRNLEX', 'NIFEXT', 'obo', 'NIFRID')
    ents = set(e for e in chain((o for _, o in g[:owl.someValuesFrom:]),
                                (o for _, o in g[:rdfs.subClassOf:]),
                                g.predicates(),
                                (s for s in g.subjects() if isinstance(s, rdflib.URIRef)
                                 and OntId(s).prefix not in bads))
               if isinstance(e, rdflib.URIRef) and not isinstance(e, rdflib.BNode)
               and (OntId(e).prefix not in bads or OntId(e).prefix in ('BIRNLEX', 'NIFEXT', 'NLX')))

    # deal with conflicting use cases for ilx terms in the ontology
    # yes, we really need perspectives working in interlex so that we
    # can have the FMA query use case and the UBERON use case separated
    ents_ilx = set(e for e in ents if OntId(e).prefix == 'ILX')
    terms_ilx = set(OntTermInterLexOnly(e) for e in ents_ilx)
    missing_sup = [t for t in terms_ilx if 'rdfs:subClassOf' not in t.predicates]
    ilx_supers = set(
        o for t in terms_ilx
        if 'rdfs:subClassOf' in t.predicates
        for o in t.predicates['rdfs:subClassOf'])
    fmaish = [t for t in terms_ilx
              if 'rdfs:subClassOf' in t.predicates and
              # wow if you leave the if out here it completely breaks
              # error message locations and syntax messages O_O
              [s for s in t.predicates['rdfs:subClassOf'] if s.prefix == 'FMA']]
    # supsub and sup_terms intentionally do not cover ilx_supers
    ilx_sup_terms = set(s.asTerm() for s in ilx_supers if s.prefix in ('ILX',))
    isupsub = set(s.u for s in ilx_supers
                  if s.prefix not in ('ILX', 'FMA',) and
                  s.prefix not in bads or
                  s.prefix in ('BIRNLEX', 'NIFEXT', 'NLX'))
    ilx_terms = terms_ilx | ilx_sup_terms
    #breakpoint()

    ents = (ents - ents_ilx) | isupsub
    #terms |= set(Async()(deferred(OntTerm)(e) for e in ents))
    terms |= set(OntTermOntologyOnly(e) for e in ents)

    all_defined_by = set(o for t in terms
                         if t('rdfs:isDefinedBy')
                         for o in t.predicates['rdfs:isDefinedBy']
                         if o.prefix not in bads)

    #fma_source = [t for t in terms if t('rdfs:subClassOf', depth=10, asTerm=True)
                  #and [o for o in t.predicates['rdfs:subClassOf'] if o.prefix == 'FMA']]
    #breakpoint()

    all_supers = set(o for t in terms
                     if t('rdfs:subClassOf', depth=10, asTerm=True)
                     for predicates in ([o for o in t.predicates['rdfs:subClassOf']
                                         # avoid pulling in multiparent cross ontology ?
                                         if o.prefix != 'FMA'],)
                     for o in (predicates if predicates else t.predicates['rdfs:subClassOf'])
                     if o.prefix not in bads)

    terms |= all_supers

    terms |= {OntTermOntologyOnly('UBERON:0002301'),  # cortical layer
              OntTermOntologyOnly('UBERON:0008933'),  # S1
              OntTermOntologyOnly('SAO:1813327414'),
              OntTermOntologyOnly('NCBITaxon:9606'),
              OntTermOntologyOnly('NCBITaxon:9685'),
              OntTermOntologyOnly('SO:0000704'),
    }

    def makeTraverse(*predicates):
        done = set()
        def traverse(term):
            if term in done:
                return term

            done.add(term)
            for predicate in predicates:
                if term(predicate, asTerm=True):
                    for superpart in term.predicates[predicate]:
                        if superpart.prefix not in ('BFO', 'NLX', 'BIRNLEX', 'NIFEXT', 'FMA'):  # continuant and occurent form a cycle >_<
                            yield superpart
                            yield from traverse(superpart)

        return traverse

    partOf = makeTraverse('partOf:', 'RO:0002433', 'rdfs:subClassOf', 'ilx.partOf:')  # 2433 CTMO

    all_sparts = set(o for t in terms  # FIXME this is broken ...
                     for o in partOf(t)
                     if o.prefix not in bads)

    terms |= all_sparts
    terms |= ilx_terms

    #sctrips = (t for T in sorted(set(asdf for t in terms for
                                     #asdf in (t, *t('rdfs:subClassOf', depth=1, asTerm=True))
                                     #if asdf.prefix not in bads), key=lambda t: (not t.label, t.label))
               #for t in T.triples('rdfs:subClassOf'))

    a = rdf.type
    def helper_triples():
        # model
        yield ilxtr.Phenotype, rdfs.subClassOf, OntId('BFO:0000016').u

        # part of for markram
        paxS1 = OntId('PAXRAT:794').u
        uS1 = OntId('UBERON:0008933').u
        yield paxS1, rdfs.subClassOf, uS1
        yield from cmb.restriction(ilxtr.isDelineatedBy, paxS1)(uS1)
        yield from cmb.restriction(ilxtr.delineates, uS1)(paxS1)

        # missing labels
        #yield OntId('CHEBI:18234').u, rdfs.label, rdflib.Literal("α,α'-trehalose 6-mycolate")  # off by one for dopamine CHEBI:18243
        yield OntId('BFO:0000050').u, rdfs.label, rdflib.Literal("part of")

        # receptor roles
        gar = ilxtr.GABAReceptor
        garole = OntId('NLXMOL:1006001').u
        yield gar, a, owl.Class
        yield gar, rdfs.label, rdflib.Literal('GABA receptor')
        yield gar, NIFRID.synonym, rdflib.Literal('GABAR')
        yield from cmb.Restriction(owl.equivalentClass)(OntId('hasRole:').u, garole)(gar)

        glr = ilxtr.glutamateReceptor
        glrole = OntId('SAO:1164727693').u
        yield glr, a, owl.Class
        yield glr, rdfs.label, rdflib.Literal('Glutamate receptor')
        yield glr, NIFRID.synonym, rdflib.Literal('GluR')
        yield from cmb.Restriction(owl.equivalentClass)(OntId('hasRole:').u, glrole)(glr)

        yield ilxtr.parcellationLabel, rdfs.subClassOf, OntId('UBERON:0001062').u

    def _old_trips(self):
        # alt fix using query neuron
        # TODO the proper way to implement this is using unionOf logical phenotype
        # since we do want to restrict it to Neurons
        n_query = ilxtr.NeuronQuery
        yield n_query, a, owl.Class
        yield n_query, rdfs.subClassOf, _NEURON_CLASS

        sst_query = ilxtr.NeuronQuerySST
        yield sst_query, a, owl.Class
        yield sst_query, rdfs.subClassOf, n_query
        yield sst_query, rdfs.label, rdflib.Literal('Somatostatin Neuron (query)')
        yield sst_query, NIFRID.synonym, rdflib.Literal('Sst Neuron (query)')
        rf = cmb.Restriction(rdf.first)
        yield from cmb.EquivalentClass(owl.unionOf)(
            *(rf(ilxtr.hasMolecularPhenotype, m) for m in sst_members),
        )(sst_query)

        pv_query = ilxtr.NeuronQueryPV
        yield pv_query, a, owl.Class
        yield pv_query, rdfs.subClassOf, n_query
        yield pv_query, rdfs.label, rdflib.Literal('Parvalbumin Neuron (query)')
        yield pv_query, NIFRID.synonym, rdflib.Literal('Pvalb Neuron (query)')

        yield from cmb.EquivalentClass(owl.unionOf)(
            *(rf(ilxtr.hasMolecularPhenotype, m) for m in pv_members),
        )(pv_query)


    class neuronUtility(Ont):
        remote_base = str(NIFRAW['neurons/'])
        path = 'ttl/'  # FIXME should be ttl/utility/ but would need the catalog file working
        filename = 'npo'
        name = 'Utility ontology for neuron development'
        imports = (NIFRAW['neurons/ttl/bridge/neuron-bridge.ttl'],
                   NIFRAW['neurons/ttl/generated/allen-transgenic-lines.ttl'],
                   #NIFRAW['neurons/ttl/generated/neurons/allen-cell-instances.ttl'],  # too slow
                   NIFRAW['neurons/ttl/generated/parcellation/mbaslim.ttl'],
                   rdflib.URIRef('http://purl.obolibrary.org/obo/bfo.owl'),

        )
        prefixes = oq.OntCuries._dict
        def _triples(self, terms=terms):
            yield from helper_triples()
            yield OntId('overlaps:').u, rdfs.label, rdflib.Literal('overlaps')
            yield ilx_partOf, rdfs.subPropertyOf, OntId('BFO:0000050').u

            yield OntId('SAO:1224657022'), rdfs.subClassOf, OntId('BFO:0000040').u  # needed for cfde

            yield OntId('BFO:0000050').u, a, owl.ObjectProperty
            yield OntId('BFO:0000050').u, a, owl.TransitiveProperty
            yield OntId('BFO:0000050').u, owl.inverseOf, OntId('BFO:0000051').u
            yield OntId('BFO:0000050').u, rdfs.subPropertyOf, OntId('overlaps:').u

            yield OntId('BFO:0000051').u, a, owl.ObjectProperty
            yield OntId('BFO:0000051').u, a, owl.TransitiveProperty
            yield OntId('BFO:0000051').u, rdfs.label, rdflib.Literal('has part')
            yield OntId('BFO:0000051').u, rdfs.subPropertyOf, OntId('overlaps:').u

            yield OntId('RO:0002433').u, rdfs.label, rdflib.Literal('contributes to morphology of')
            yield OntId('RO:0002433').u, rdfs.subPropertyOf, OntId('overlaps:').u

            done = [
                # these don't have labels in the ontology and they are cells which is bad
                # so skip them by putting them in done from the start
                OntTermOntologyOnly('NIFSTD:BAMSC1121'),
                OntTermOntologyOnly('NIFSTD:BAMSC1126'),
            ]
            _ipo_done = []
            cortical_layer = OntTermOntologyOnly('UBERON:0002301')
            #yield from cortical_layer.triples_simple
            #yield from cortical_layer('hasPart:')

            if not terms:
                raise BaseException('WHAT')

            skip = 'TEMPIND', 'npokb', 'mmset'
            while terms:
                next_terms = []
                for term in terms:
                    if term in done:
                        continue

                    done.append(term)
                    if term.curie and [p for p in skip if term.curie.startswith(p)]:
                        continue

                    if term.label is not None and not isinstance(term.label, str):
                        log.warning(f'bad label? {term}')  # next line will error

                    if (not term.label or (not term.label.lower().endswith('neuron') and
                                           not term.label.lower().endswith('cell') and
                                           not term.label.lower().endswith('cell outer'))
                        or term.URIRef == _NEURON_CLASS):
                        haveLabel = False
                        for s, p, o in term.triples_simple:
                            if p == rdfs.label:
                                haveLabel = True
                            if s == definition and p == rdf.type:  # FIXME :/
                                o = owl.AnnotationProperty

                            yield s, p, o
                            #log.debug(f'{o!r}')
                            if isinstance(o, rdflib.URIRef):
                                if o == owl.Restriction:
                                    continue

                                no = OntTermOntologyOnly(o)
                                if no not in done:
                                    if term.prefix not in bads and no.prefix in bads:
                                        if not no.label:
                                            continue
                                        try:
                                            gen = OntTermOntologyOnly.query(term=no.label)
                                            noo = next(gen)
                                            if noo.curie in ('owl:Class', 'owl:Thing'):
                                                continue
                                            elif noo != no and noo.prefix not in bads:
                                                no = noo
                                            else:
                                                continue
                                        except StopIteration:
                                            continue

                                    next_terms.append(no)

                        if not haveLabel:
                            ot = OntTerm(term)  # include InterLex
                            if ot.label:
                                yield ot.u, rdfs.label, rdflib.Literal(ot.label)
                                if hasattr(ot, '_graph'):
                                    ipo = str(ilx_partOf)
                                    for s, p, o in ot._graph.subjectGraphClosure(ot.u):
                                        yield s, p, o
                                        if str(p) == ipo and isinstance(o, rdflib.URIRef):
                                            if s == o:
                                                pass  # prevent partOf self from creeping in via scigraph
                                            elif (s, o) not in _ipo_done:
                                                _ipo_done.append((s, o))
                                                yield from cmb.restriction(
                                                    BFO_partOf, o)(s)

                                                no = OntTermOntologyOnly(o)
                                                if no not in done:
                                                    next_terms.append(no)

                terms = next_terms


    class neuronUtilityBig(Ont):
        remote_base = str(NIFRAW['neurons/'])
        path = 'ttl/'  # FIXME should be ttl/utility/ but would need the catalog file working
        filename = 'npo-large'
        name = 'Utility ontology for neuron development'
        imports = (NIFRAW['neurons/ttl/bridge/neuron-bridge.ttl'],
                   NIFRAW['neurons/ttl/bridge/chemical-bridge.ttl'],
                   NIFRAW['neurons/ttl/generated/parcellation.ttl'],
                   NIFRAW['neurons/ttl/generated/parcellation-artifacts.ttl'],
                   NIFRAW['neurons/ttl/generated/uberon-parcellation-mappings.ttl'],
                   NIFRAW['neurons/ttl/generated/parcellation/mbaslim.ttl'],
                   NIFRAW['neurons/ttl/generated/parcellation/paxinos-rat-labels.ttl'],
                   rdflib.URIRef('http://purl.obolibrary.org/obo/uberon.owl'),
        )
        prefixes = oq.OntCuries._dict


        def _triples(self):
            yield from helper_triples()


    CustomTurtleSerializer.roundtrip_prefixes = ('',)
    nu, nub, *_ = build(neuronUtility, neuronUtilityBig, n_jobs=1)
    CustomTurtleSerializer.roundtrip_prefixes = True


def main():
    from docopt import docopt
    args = docopt(__doc__)
    dep = args['dep']
    all = args['all']
    old = args['old']               or all
    release = args['release']       or all
    phenotypes = args['phenotypes'] or all or release or old
    bridge = args['bridge']         or all
    models = args['models']         or all or release
    dev = args['dev']               or all or release
    indicators = args['indicators'] or all or release
    sheets = args['sheets']         or all or release

    if dep:
        from neurondm.lang import Config
        from neurondm.compiled.common_usage_types import config as c_config
        cnrns = c_config.neurons()
        config = Config('cut-development-fixed')
        nns = [n.asUndeprecated() for n in cnrns]
        config.write()
        config.write_python()
        breakpoint()
        return

    if indicators:
        from neurondm import indicators as ind
        ind.main()

    if phenotypes:
        syn_mappings, pedge, ilx_start, phenotypes, defined_graph = make_phenotypes()
        if old:
            syn_mappings['thalamus'] = defined_graph.expand('UBERON:0001879')
            expand_syns(syn_mappings)

    if models:
        make_models()

    if old:
        ilx_start = make_neurons(syn_mappings, pedge, ilx_start, defined_graph, old=old)

    if sheets:
        from neurondm import sheets
        sheets.main()

    if dev:
        make_devel()

    if bridge:
        make_bridge()

    if __name__ == '__main__':
        #breakpoint()
        pass

if __name__ == '__main__':
    main()



import unittest
from .common import skipif_no_net


@skipif_no_net
class TestWrite(unittest.TestCase):
    def test_load_huang(self):
        from neurondm import Config
        # FIXME placeholder for loading and roundtripping
        # neurons with other neurons as asserted equivalent
        # or disjoint classes
        config = Config('huang-2017')
        config.load_existing()



import re
from collections import defaultdict
from pyontutils.sheets import Sheet
from pyontutils.namespaces import ilxtr, TEMP, rdfs, skos, owl
from neurondm.core import Config, NeuronEBM, Phenotype, EntailedPhenotype, log, OntId


class NeuronApinatSimple(NeuronEBM):
    owlClass = ilxtr.NeuronApinatSimple
    shortname = 'apinat'


class AtN(Sheet):
    name = 'apinat-to-npo-sue'


sheet_classes = [
    type(f'AtN{sname.replace(" ", "_")}',
         (AtN,), dict(sheet_name=sname))
    for sname in (
            'aacar',
            'splen',
            'sstom',
            'sdcol',

            'bromo',
            'kblad',
            'bolew',
            'pancr',
    )]


def map_predicates(sheet_pred):
    p = {
        '': TEMP.BROKEN_EMPTY,
        'Soma-Location': ilxtr.hasSomaLocatedIn,
        'Axon-Terminal-Location': ilxtr.hasAxonPresynapticElementIn,
        'Axon-Location': ilxtr.hasAxonLocatedIn,
        'Dendrite-Location': ilxtr.hasDendriteLocatedIn,
        #'Dendrite-Terminal-Location': ilxtr.hasDendriteSensorySubcellularElementIn,
        'Forward-Connection': ilxtr.hasForwardConnectionPhenotype,
        'Axon-Sensory-Location': ilxtr.hasAxonSensorySubcellularElementIn,
        'Equivalent-To': owl.equivalentClass,
        'Functional-Circuit-Role': ilxtr.hasFunctionalCircuitRolePhenotype,
        'entail:hasInstanceInTaxon': ilxtr.hasInstanceInTaxon,
        'Organ-Destination': ilxtr.hasOrganTarget,
        'Has-Phenotype': ilxtr.hasPhenotype,  # needed for aacar 7 and 8
        'Projection-Phenotype': ilxtr.hasProjectionPhenotype,
    }[sheet_pred]
    return p


def main():
    cs = [c() for c in sheet_classes]
    trips = [[cl] + [c.value for c in
                     (r.neuron_id(), r.exact_location(), r.location_id())]
             for cl in cs for r in cl.rows()
             if r.row_index > 0 and r.neuron_id().value
             and (not hasattr(r, 'exclude') or not r.exclude().value)]

    to_add = []
    dd = defaultdict(list)
    for c, _s, _p, _o in trips:
        for x in (_s, _p, _o):
            if re.match(r'(^[\s]+[^\s].*|.*[^\s][\s]+$)', x):
                msg = f'leading/trailing whitespace in {c} {_s!r} {_p!r} {_o!r}'
                log.error(msg)
                raise ValueError(msg)

        s = OntId(_s)
        try:
            p = map_predicates(_p)
        except KeyError as e:
            log.error(f'sigh {s}')
            raise e
        o = OntId(_o)

        if p == owl.equivalentClass or p == ilxtr.hasOrganTarget:
            to_add.append((s.u, p, o.u))
            continue

        try:
            if _p.startswith('entail:'):  # XXX FIXME hack
                dd[s].append(EntailedPhenotype(o, p))
            else:
                dd[s].append(Phenotype(o, p))
        except TypeError as e:
            log.error(f'bad data for {c} {s} {p} {o}')
            raise e

    problems = ('8a', '8v', 'sstom-6', 'keast-2', 'sdcol-k', 'sdcol-l')
    def eff(n):
        return bool([x for x in problems
                     if x in n.id_ and '20' not in n.id_])

    config = Config('apinat-simple-sheet')
    sigh = []
    nrns = []
    for id, phenos in dd.items():
        n = NeuronApinatSimple(*phenos, id_=id)
        if eff(n):
            n._sigh()  # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX FIXME figure out why this is not getting called internally
            sigh.append(n)

        nrns.append(n)

    config.write()
    labels = (
        ilxtr.genLabel, ilxtr.localLabel, ilxtr.simpleLabel,
        ilxtr.simpleLocalLabel, rdfs.label, skos.prefLabel)
    to_remove = [t for t in config._written_graph if t[1] in labels]
    [config._written_graph.remove(t) for t in to_remove]
    [config._written_graph.add(t) for t in to_add]
    config._written_graph.write()
    config.write_python()
    return config,


if __name__ == '__main__':
    main()



import unittest
from neurondm import OntTerm as OntTerm_
from neurondm.core import OntTermOntologyOnly
from pyontutils.core import OntGraph
from pyontutils.namespaces import partOf
from .common import skipif_no_net


@skipif_no_net
class TriplesSimple:
    OntTerm = None
    def test_part_of(self):
        eeeee = self.OntTerm('UBERON:0008933',
                             label='primary somatosensory cortex')
        g = OntGraph()
        [g.add(t) for t in eeeee.triples_simple]
        g.debug()
        po = [t for t in eeeee.triples_simple if partOf in t]
        assert po, 'sadness'


class TestOT(TriplesSimple, unittest.TestCase):
    OntTerm = OntTerm_


class TestOTOO(TriplesSimple, unittest.TestCase):
    OntTerm = OntTermOntologyOnly



import os
import rdflib
from pyontutils.core import OntGraph, OntResIri, OntResPath
from pyontutils.namespaces import rdfs, ilxtr
from neurondm.core import Config, graphBase, log
from neurondm.core import OntTerm, OntId, RDFL
from neurondm import orders


def multi_orig_dest(neuron):
    for dim in neuron.edges:
        if 'hasAxonPre' in dim or 'hasAxonSens' in dim or 'hasSoma' in dim:
            objs = list(neuron.getObjects(dim))
            if len(objs) > 1:
                return True


def makelpesrdf():
    collect = []
    def lpes(neuron, predicate):
        """ get predicates from python bags """
        # TODO could add expected cardinality here if needed
        return [str(o) for o in neuron.getObjects(predicate)
                if not collect.append((predicate, o))]

    def lrdf(neuron, predicate):
        """ get predicates from graph """
        return [  # XXX FIXME core_graph bad etc.
            str(o) for o in
            neuron.core_graph[neuron.identifier:predicate]]

    return lpes, lrdf, collect


def simplify(e):
    if e is None:
        return
    elif isinstance(e, rdflib.Literal):  # blank case
        return e.toPython()
    else:
        return OntTerm(e).curie


def simplify_nested(f, nested):
    for e in nested:
        if isinstance(e, list) or isinstance(e, tuple):
            yield tuple(simplify_nested(f, e))
        elif isinstance(e, orders.rl):
            yield orders.rl(f(e.region), f(e.layer))
        else:
            yield f(e)


def for_composer(n, cull=False):
    lpes, lrdf, collect = makelpesrdf()
    _po = n.partialOrder()
    fc = dict(
        id = str(n.id_),
        label = str(n.origLabel),
        origin = lpes(n, ilxtr.hasSomaLocatedIn),
        dest = (
            # XXX looking at this there seems to be a fault assumption that
            # there is only a single destination type per statement, this is
            # not the case, there is destination type per destination
            [dict(loc=l, type='AXON-T') for l in lpes(n, ilxtr.hasAxonPresynapticElementIn)] +
            # XXX I strongly reccoment renaming this to SENSORY-T so that the
            # short forms are harder to confuse A-T and S-T
            [dict(loc=l, type='AFFERENT-T') for l in lpes(n, ilxtr.hasAxonSensorySubcellularElementIn)]
        ),
        order = tuple(simplify_nested(simplify, _po)) if _po else [],
        path = (  # TODO pull ordering from partial orders (not implemented in core atm)
            [dict(loc=l, type='AXON') for l in lpes(n, ilxtr.hasAxonLocatedIn)] +
            # XXX dendrites don't really ... via ... they are all both terminal and via at the same time ...
            [dict(loc=l, type='DENDRITE') for l in lpes(n, ilxtr.hasDendriteLocatedIn)]
        ),
        #laterality = lpes(n, ilxtr.hasLaterality),  # left/rigth tricky ?
        #projection_laterality = lpes(n, ilxtr.???),  # axon located in contra ?
        species =            lpes(n, ilxtr.hasInstanceInTaxon),
        sex =                lpes(n, ilxtr.hasBiologicalSex),
        circuit_type =       lpes(n, ilxtr.hasCircuitRolePhenotype),
        phenotype =          lpes(n, ilxtr.hasAnatomicalSystemPhenotype),  # current meaning of composer phenotype
        anatomical_system =  lpes(n, ilxtr.hasAnatomicalSystemPhenotype),
        # there are a number of dimensions that we aren't converting right now
        dont_know_fcrp =     lpes(n, ilxtr.hasFunctionalCircuitRolePhenotype),
        other_phenotype = (  lpes(n, ilxtr.hasPhenotype)
                           + lpes(n, ilxtr.hasMolecularPhenotype)
                           + lpes(n, ilxtr.hasProjectionPhenotype)),
        forward_connection = lpes(n, ilxtr.hasForwardConnectionPhenotype),

        # direct references from individual individual neurons
        provenance =      lrdf(n, ilxtr.literatureCitation),
        sentence_number = lrdf(n, ilxtr.sentenceNumber),
        note_alert =      lrdf(n, ilxtr.alertNote),
        # XXX provenance from ApiNATOMY models as a whole is not ingested
        # right now because composer lacks support for 1:n from neuron to
        # prov, (or rather lacks prov collections) and because it attaches
        # prov to the sentece, which does not exist for all neurons

        # TODO more ...
        # notes = ?

        # for _ignore, hasClassificationPhenotype is used for ApiNATOMY
        # unlikely to be encountered for real neurons any time soon
        _ignore = lpes(n, ilxtr.hasClassificationPhenotype),  # used to ensure we account for all phenotypes
    )
    npo = set((p.e, p.p) for p in n.pes)
    cpo = set(collect)
    unaccounted_pos = npo - cpo
    if unaccounted_pos:
        log.warning(
            (n.id_, [[n.in_graph.namespace_manager.qname(e) for e in pos]
                     for pos in unaccounted_pos]))
    return {k:v for k, v in fc.items() if v} if cull else fc


def location_summary(neurons, services, anatent_simple=False):
    import csv
    OntTerm.query._services = services
    locations = sorted(set(
        OntTerm(pe.p) for n in neurons for pe in n.pes
        if pe.e in n._location_predicates))
    [_.fetch() for _ in locations]

    _loc_src = sorted(set(
        (OntTerm(pe.p), n.id_) for n in neurons for pe in n.pes
        if pe.e in n._location_predicates))

    ls2 = dict()
    for _l, _s in _loc_src:
        if _l not in ls2:
            ls2[_l] = []

        ls2[_l].append(_s)

    loc_src = dict()
    for k, v in ls2.items():
        nlp = [iri for iri in v if '/sparc-nlp/' in iri]
        apinat = [iri for iri in v if '/neuron-type-' in iri]
        both = apinat and nlp
        source = 'both' if both else 'nlp' if nlp else 'apinat' if apinat else 'unknown'
        loc_src[k] = source

    def key(t):
        return (t.prefix, t.label[0].lower()
                if isinstance(t, tuple)
                else t.lower())

    if anatent_simple:
        header = 'label', 'curie', 'iri', 'source'
        rows = (
            [header] +
            [(_.label, _.curie, _.iri, loc_src[_]) for _ in sorted(locations, key=key)])
        with open('/tmp/npo-nlp-apinat-location-summary.csv', 'wt') as f:
            csv.writer(f, lineterminator='\n').writerows(rows)

    else:
        header = 'o', 'o_label', 'o_synonym'
        rows = (
            [header] +
            [(_.iri, _.label, syn) for _ in sorted(locations, key=key)
             for syn in _.synonyms])
        with open('/tmp/anatomical_entities.csv', 'wt') as f:
            csv.writer(f, lineterminator='\n').writerows(rows)


def main(local=False, anatomical_entities=False, anatent_simple=False):
    # if (local := True, anatomical_entities := True, anatent_simple := False):

    config = Config('random-merge')
    g = OntGraph()  # load and query graph

    # remove scigraph and interlex calls
    graphBase._sgv = None
    del graphBase._sgv
    if len(OntTerm.query._services) > 1:
        # backup services and avoid issues on rerun
        _old_query_services = OntTerm.query._services
        _noloc_query_services = _old_query_services[1:]

    OntTerm.query._services = (RDFL(g, OntId),)

    # base paths to ontology files
    gen_neurons_path = 'ttl/generated/neurons/'
    suffix = '.ttl'
    if local:
        from pyontutils.config import auth
        olr = auth.get_path('ontology-local-repo')
        local_base = olr / gen_neurons_path
    else:
        orr = 'https://raw.githubusercontent.com/SciCrunch/NIF-Ontology/neurons/'
        remote_base = orr + gen_neurons_path

    # full imports
    for f in ('apinat-partial-orders',
              'apinat-pops-more',
              'apinat-simple-sheet',
              'sparc-nlp'):
        if local:
            ori = OntResPath(local_base / (f + suffix))
        else:
            ori = OntResIri(remote_base + f + suffix)
        [g.add(t) for t in ori.graph]

    # label only imports
    for f in ('apinatomy-neuron-populations',
              '../../npo'):
        p = os.path.normpath(gen_neurons_path + f)
        if local:
            ori = OntResPath(olr / (p + suffix))
        else:
            ori = OntResIri(orr + p + suffix)

        [g.add((s, rdfs.label, o)) for s, o in ori.graph[:rdfs.label:]]

    config.load_existing(g)
    neurons = config.neurons()  # scigraph required here if deps not removed above

    # ingest to composer starts here
    mvp_ingest = [n for n in neurons if not multi_orig_dest(n)]

    dims = set(p for n in neurons for p in n.edges)  # for reference
    fcs = [for_composer(n) for n in mvp_ingest]
    _fcne = [for_composer(n, cull=True) for n in mvp_ingest]  # exclude empties for easier manual review

    # example neuron
    n = mvp_ingest[0]
    fc = for_composer(n)

    if anatomical_entities:
        location_summary(neurons, _noloc_query_services, anatent_simple)


if __name__ == '__main__':
    main()



import re
from collections import defaultdict
import rdflib
from pyontutils.sheets import Sheet
from pyontutils.namespaces import ilxtr, TEMP, rdfs, skos, owl, interlex_namespace
from neurondm.core import Config, NeuronEBM, Phenotype, NegPhenotype, log, OntId, add_partial_orders
from neurondm import orders


class NeuronSparcNlp(NeuronEBM):
    owlClass = ilxtr.NeuronSparcNlp
    shortname = 'sprcnlp'


class NLP1(Sheet):
    name = 'off-nlp-1'


class NLP2(Sheet):
    name = 'off-nlp-2'


def nlp_ns(name):
    return rdflib.Namespace(interlex_namespace(f'tgbugs/uris/readable/sparc-nlp/{name}/'))


snames = {
    'MM Set 1': (NLP1, nlp_ns('mmset1')),
    'MM Set 2 Cranial Nerves': (NLP1, nlp_ns('mmset2cn')),
    'MM Set 4': (NLP2, nlp_ns('mmset4')),
}


sheet_classes = [
    type(f'{base.__name__}{sname.replace(" ", "_")}',
         (base,), dict(sheet_name=sname))
    for sname, (base, ns) in snames.items()]


def map_predicates(sheet_pred):
    p = {
        '': TEMP.BROKEN_EMPTY,
        'Soma': ilxtr.hasSomaLocatedIn,
        'Axon terminal': ilxtr.hasAxonPresynapticElementIn,
        'Axon': ilxtr.hasAxonLocatedIn,
        'Dendrite': ilxtr.hasDendriteLocatedIn,
        'Axon sensory terminal': ilxtr.hasAxonSensorySubcellularElementIn,
        'hasInstanceInTaxon': ilxtr.hasInstanceInTaxon,
        'hasPhenotype': ilxtr.hasPhenotype,
        'hasAnatomicalSystemPhenotype': ilxtr.hasAnatomicalSystemPhenotype,
        'hasBiologicalSex': ilxtr.hasBiologicalSex,
        'hasCircuitRolePhenotype': ilxtr.hasCircuitRolePhenotype,
        'hasForwardConnectionPhenotype': ilxtr.hasForwardConnectionPhenotype,  # FIXME this needs to be unionOf
    }[sheet_pred]
    return p


def ind_to_adj(ind_uri):
    inds = sorted(set([i for i, u in ind_uri]))
    edges = []
    for a, b in zip(inds[:-1], inds[1:]):
        for i, iu in ind_uri:
            # XXX yes this is a horribly inefficient way to do this
            # index them in a dict if it becomes an issue
            if i == a:
                for j, ju in ind_uri:
                    if j == b:
                        edge =  iu, ju
                        edges.append(edge)

    return edges


def main():
    cs = [c() for c in sheet_classes]
    trips = [[cl] + [c.value for c in
                     (r.id(), r.relationship(), r.identifier())]
             for cl in cs for r in cl.rows()
             if r.row_index > 0 and r.id().value
             #and (not hasattr(r, 'exclude') or not r.exclude().value)
             and r.proposed_action().value.lower() != "don't add"
             ]

    to_add = []

    def vl(meth):
        return rdflib.Literal(meth().value)

    def asdf(s, p, rm):
        v = vl(rm)
        if v:
            # XXX watch out for non-homogenous subject types
            # (e.g. OntId vs rdflib.URIRef)
            to_add.append((s.u, p, v))

    dd = defaultdict(list)
    ec = {}
    for cl in cs:
        _, nlpns = snames[cl.sheet_name]
        for r in cl.rows():
            if (r.row_index > 0 and
                r.id().value and
                r.proposed_action().value.lower() != "don't add"):
                # extra trips
                #print(repr(r.id()))
                s = OntId(nlpns[r.id().value])
                #print(s)
                asdf(s, ilxtr.sentenceNumber, r.sentence_number)
                asdf(s, ilxtr.curatorNote, r.different_from_existing)
                asdf(s, ilxtr.curatorNote, r.curation_notes)
                asdf(s, ilxtr.reviewNote, r.review_notes)
                asdf(s, ilxtr.reference, r.reference_pubmed_id__doi_or_text)
                asdf(s, ilxtr.literatureCitation, r.literature_citation)
                asdf(s, rdfs.label, r.neuron_population_label_a_to_b_via_c)
                if hasattr(r, 'alert_explanation'):
                    asdf(s, ilxtr.alertNote, r.alert_explanation)

                if hasattr(r, 'explicit_complement') and r.explicit_complement().value:
                    p = map_predicates(r.relationship().value)
                    o = OntId(r.explicit_complement().value)
                    ec[(s, p)] = o

                if hasattr(r, 'axonal_course_poset') and r.axonal_course_poset().value:
                    # s.u and OntId(...).u to avoid duplicate subjects/objects in the graph
                    # due to type vs instance issues for rdflib.URIRef and OntId
                    dd[s.u].append((int(r.axonal_course_poset().value), OntId(r.identifier().value).u))

    sorders = {k:sorted(v) for k, v in dd.items()}
    sadj = {k:ind_to_adj(v) for k, v in sorders.items()}
    snst = {k:orders.adj_to_nst(v) for k, v in sadj.items()}

    dd = defaultdict(list)
    for c, _s, _p, _o in trips:
        _, nlpns = snames[c.sheet_name]
        for x in (_s, _p, _o):
            if re.match(r'(^[\s]+[^\s].*|.*[^\s][\s]+$)', x):
                msg = f'leading/trailing whitespace in {c} {_s!r} {_p!r} {_o!r}'
                log.error(msg)
                raise ValueError(msg)

        s = OntId(nlpns[_s])
        try:
            p = map_predicates(_p)
        except KeyError as e:
            log.error(f'sigh {s}')
            raise e

        if p == ilxtr.hasForwardConnectionPhenotype:
            _o = nlpns[_o]

        o = OntId(_o)

        if p == owl.equivalentClass:
            to_add.append((s.u, p, o.u))
            continue

        try:
            dd[s].append(Phenotype(o, p))
        except TypeError as e:
            log.error(f'bad data for {c} {s} {p} {o}')
            raise e

        if ec and (s, p) in ec:
            dd[s].append(NegPhenotype(ec[(s, p)], p))


    config = Config('sparc-nlp')

    sigh = []
    nrns = []
    for id, phenos in dd.items():
        n = NeuronSparcNlp(*phenos, id_=id)
        if False and eff(n):
            n._sigh()  # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX FIXME figure out why this is not getting called internally
            sigh.append(n)

        nrns.append(n)

    Phenotype.in_graph.bind('mmset4', snames['MM Set 4'][-1])  # XXX FIXME UGH
    config.write()
    labels = (
        #ilxtr.genLabel,
        ilxtr.localLabel, ilxtr.simpleLabel,
        ilxtr.simpleLocalLabel, rdfs.label, skos.prefLabel)
    to_remove = [t for t in config._written_graph if t[1] in labels]
    [config._written_graph.remove(t) for t in to_remove]
    [config._written_graph.add(t) for t in to_add]
    add_partial_orders(config._written_graph, snst)
    # uncomment to debug type vs instance issues
    #sigh = sorted(set(s for s in config._written_graph.subjects() if isinstance(s, rdflib.URIRef)))
    config._written_graph.write()
    config.write_python()
    return config,


if __name__ == '__main__':
    main()



import os
import sys
import unittest
from pathlib import Path
from tempfile import gettempdir
import pytest
from git import Repo
from neurondm.core import log

log.setLevel('DEBUG')

SKIP_NETWORK = ('SKIP_NETWORK' in os.environ or
                'FEATURES' in os.environ and 'network-sandbox' in os.environ['FEATURES'])
skipif_no_net = pytest.mark.skipif(SKIP_NETWORK, reason='Skipping due to network requirement')

testing_base = Path(gettempdir(), f'.neurons-testing-base-{os.getpid()}')
pyel = testing_base / 'compiled'
tel = testing_base


@skipif_no_net
class _TestNeuronsBase(unittest.TestCase):
    def setUp(self):
        if not pyel.exists():
            pyel.mkdir(parents=True)  # recrusive clean is called after every test
            (pyel / '__init__.py').touch()

        repo = Repo.init(testing_base.as_posix())

    def tearDown(self):
        def recursive_clean(path):
            for thing in path.iterdir():
                if thing.is_dir():
                    recursive_clean(thing)
                else:
                    thing.unlink()  # will rm the file

            path.rmdir()

        path = testing_base
        if path.exists():
            recursive_clean(path)



import unittest
import rdflib
from neurondm.core import (LabelMaker,
                           Phenotype,
                           NegPhenotype,
                           Neuron,
                           LogicalPhenotype,
                           AND,
                           OR,)


class TestLabelMaker(unittest.TestCase):

    def setUp(self):
        self.lm = LabelMaker()

        def ns(suffix):
            if isinstance(suffix, rdflib.URIRef):
                raise TypeError('you called this twice')
            return self.lm.predicate_namespace[suffix]

        def ms(*suffixes, p=None):
            return Neuron(*[Phenotype(s, p) for s in suffixes])

        def ls(*suffixes, p=None, op=AND):
            # not reasonable but useful for display testing
            return Neuron(LogicalPhenotype(
                op,
                *[Phenotype(s, p) for s in suffixes],
                *[NegPhenotype(s, p) for s in suffixes]))

        self.ns = ns
        self.ms = ms
        self.ls = ls

    def test_logical(self):
        ns = self.ns
        ls = self.ls
        hmp = ns('hasMolecularPhenotype')
        a = 'PR:000015665'  # sst
        b = 'NCBIGene:19293'  # pv
        c = 'PR:000017299'  # vip

        bads = []
        for op in (AND, OR):
            hrm = ls(a, b, c, p=hmp, op=op)
            test = hrm.label
            if not ('+PV' in test and '-VIP' in test):
                bads.append(test)

        assert not bads, bads

    def test_hasCircuitRole(self):
        ns = self.ns
        ms = self.ms
        hcr = ns('hasCircuitRolePhenotype')
        inter = ns('InterneuronPhenotype')
        intrin = ns('IntrinsicPhenotype')
        proj = ns('ProjectionPhenotype')
        prin = ns('PrincipalPhenotype')
        s = ns('SensoryPhenotype')
        m = ns('MotorPhenotype')

        sets = (
            ms(intrin, p=hcr),
            ms(inter, p=hcr),
            ms(prin, p=hcr),
            ms(proj, p=hcr),
            ms(s, p=hcr),
            ms(m, p=hcr),
            ms(inter, intrin, p=hcr),
            ms(intrin, inter, p=hcr),
        )
        wat = next(self.lm.hasCircuitRolePhenotype((Phenotype(inter),)))
        assert wat != 'intrinsic'
        for neuron in sets:
            label = self.lm(neuron)
            print(neuron.pes, label)




#!/usr/bin/env python3
""" Converts owl or ttl or raw rdflib graph into a pandas DataFrame. Saved in .pickle format.

Usage:
    allen_cell_types [options]

Options:
    -h --help                   Display this help message
    -v --version                Current version of file
    -r --refresh                Update local copy
    -i --input=<path>           Local copy of Allen Brain Atlas meta data [default: /tmp/allen-cell-types.json]
    -o --output=<path>          Output path of picklized pandas DataFrame [default: allen-cell-types]
"""
import re
import json
from pathlib import Path
import rdflib
import requests
from pyontutils.utils import TermColors as tc, relative_path
from pyontutils.core import simpleOnt, OntGraph
from pyontutils.namespaces import makePrefixes, ilxtr, definition
from pyontutils.namespaces import rdf, rdfs, owl, AIBSSPEC, npokb
from pyontutils.combinators import annotation, allDifferent, distinctMembers
from neurondm.core import auth
from neurondm.lang import *
from neurondm.simple import Cell
from .allen_type_specimen_mapping import ts_mapping
from docopt import docopt, parse_defaults


class NeuronACT(NeuronEBM):
    owlClass = ilxtr.NeuronACT
    shortname = 'AllenCT'


class AllenCellTypes:

    branch = auth.get('neurons-branch')

    prefixes = makePrefixes('definition', 'ilxtr', 'owl', 'AIBSTL', 'JAX', 'MMRRC', 'AIBSSPEC')

    def __init__(self, input, name):
        self.name = name
        self.ns = {k:rdflib.Namespace(v) for k, v in self.prefixes.items()}
        self.neuron_data = input
        self.tag_names = set()
        # self.sample_neuron()

    def avoid_url_conversion(self, string):
        if not string:
            return string
        return re.sub("/| |\(", '_', string).replace(')', '')

    def sample_neuron(self,):
        Neuron(
            Phenotype('ilxtr:apical',
                      'ilxtr:hasPhenotype',
                      label='apical - truncated'),
            Phenotype('JAX:12345',
                      'ilxtr:hasExperimentalPhenotype',
                      label='prefix+stock_number'),
        )
        print(graphBase.ttl())

    def cell_phenotypes(self, cell_specimen):
        cell_mappings = {
            'hemisphere': 'ilxtr:hasSomaLocationLaterality',
            # 'name': 'ilxtr:hasPhenotype',
        }
        phenotypes = []
        for name, value in cell_specimen.items():
            mapping = cell_mappings.get(name)
            if mapping and value:
                if name == 'hemisphere':
                    if value.lower() == 'left':
                        curie = 'UBERON:0002812'
                    elif value.lower() == 'right':
                        curie = 'UBERON:0002813'
                    else:
                        raise ValueError('got stuck with unkown hemisphere ' + value)
                phenotypes.append(
                    Phenotype(
                        curie,
                        mapping,
                    )
                )
        return phenotypes

    # TODO: wrong phenotype
    def structure_phenotypes(self, cell_specimen):
        struc = cell_specimen['structure']
        phenotypes = []
        acronym = self.avoid_url_conversion(struc['acronym'])
        curie = 'MBA:' + str(struc['id'])
        if struc:
            phenotypes.append(
                Phenotype(
                    curie,
                    'ilxtr:hasSomaLocatedIn',
                    label=acronym
                ),
            )
        return phenotypes

    def donor_phenotypes(self, cell_specimen):
        donor_mappings = {
            'sex_full_name': 'ilxtr:hasBiologicalSex'
        }
        phenotypes = []
        for name, value in cell_specimen['donor'].items():
            mapping = donor_mappings.get(name)
            if mapping and value:
                if name == 'sex_full_name':
                    if value.lower() == 'female':
                        curie = 'PATO:0000383'
                    elif value.lower() == 'male':
                        curie = 'PATO:0000384'
                    else:
                        raise ValueError('unkown sex ' + str(value))
                phenotypes.append(
                    Phenotype(
                        curie,
                        mapping,
                    ),
                )
        return phenotypes

    # TODO: Figure how to add: description, name and type
    def transgenic_lines_phenotypes(self, cell_specimen):
        transgenic_mappings = {
        }
        phenotypes = []
        for tl in cell_specimen['donor']['transgenic_lines']:
            prefix = tl['transgenic_line_source_name']
            suffix = tl['stock_number'] if tl['stock_number'] else str(tl['id'])
            name = self.avoid_url_conversion(tl['name'])
            _type = tl['transgenic_line_type_name']
            if _type == 'driver':
                if 'CreERT2' in name:  # FIXME from structured instead of name?
                    pred = ilxtr.hasDriverExpressionInducedPhenotype
                else:
                    pred = 'ilxtr:hasDriverExpressionPhenotype'
            elif _type == 'reporter':
                pred = 'ilxtr:hasReporterExpressionPhenotype'
            else:
                pred = 'ilxtr:hasExpressionPhenotype'

            line_names = []
            if prefix and suffix and prefix in ['AIBS', 'MMRRC', 'JAX']:
                if prefix == 'AIBS':
                    prefix = 'AIBSTL'
                iri = self.ns[prefix][suffix]
                phenotypes.append(Phenotype(iri, pred))
        return phenotypes

    # TODO: search if description exists
    # TODO: Create mapping for all possible types
    # TODO: Fork negatives to NegPhenotype
    def specimen_tags_phenotypes(self, cell_specimen):
        pred = 'ilxtr:hasDendriteMorphologicalPhenotype'
        specimen_tag_mappings = {
            'spiny': Phenotype('ilxtr:SpinyPhenotype', pred),
            'aspiny': NegPhenotype('ilxtr:SpinyPhenotype', pred),
            'sparsely spiny': LogicalPhenotype(AND,
                                               Phenotype('ilxtr:SpinyPhenotype', pred),
                                               Phenotype('PATO:0001609', 'ilxtr:hasPhenotypeModifier')),
            'apicalIntact': Phenotype('ilxtr:ApicalDendritePhenotype', 'ilxtr:hasMorphologicalPhenotype'),
            'apicalTruncated': LogicalPhenotype(AND,
                                                Phenotype('ilxtr:ApicalDendritePhenotype', 'ilxtr:hasMorphologicalPhenotype'),
                                                Phenotype('PATO:0000936', 'ilxtr:hasPhenotypeModifier')),
            'apicalNa': NegPhenotype('ilxtr:ApicalDendritePhenotype', 'ilxtr:hasMorphologicalPhenotype'),  # NA means there was no apical dendrite
        }
        phenotypes = []
        for tag in cell_specimen['specimen_tags']:
            if 'dendrite type' in tag['name']:
                one_two = tag['name'].split(' - ')[1]
                #if ' ' in one_two:
                    #one, two = one_two.split(' ')
                    #name = one + two.capitalize()
                #else:
                name = one_two
            else:
                one, two = tag['name'].split(' - ')
                #if two == 'NA':  # apical - NA
                    #continue
                name = one + two.capitalize()

            self.tag_names.add(tag['name'])
            # if phenotype == '+':
            if name not in specimen_tag_mappings:
                raise ValueError(name)

            phenotypes.append(specimen_tag_mappings[name]
                              if name in specimen_tag_mappings else
                              Phenotype('ilxtr:' + name, pred))
            # elif phenotype == '-': phenotypes.append(NegPhenotype(...))

        return phenotypes

    # TODO: check to see if specimen_id is really the priority
    def cell_soma_locations_phenotypes(self, cell_specimen):
        cell_soma_mappings = {
        }
        phenotypes = []
        for csl in cell_specimen['cell_soma_locations']:
            location = csl['id']
            phenotypes.append(
                Phenotype(
                    'ilxtr:' + str(location),
                    'ilxtr:hasSomaLocatedIn',
                )
            )
        return phenotypes

    def add_mouse_lineage(self, cell_specimen):
        phenotypes = [Phenotype('NCBITaxon:10090', 'ilxtr:hasInstanceInTaxon')]
        return phenotypes

    def build_phenotypes(self, cell_specimen):
        phenotype_functions = [
            self.cell_phenotypes,
            self.structure_phenotypes,
            self.donor_phenotypes,
            self.transgenic_lines_phenotypes,
            self.specimen_tags_phenotypes,
            self.add_mouse_lineage,
            # self.cell_soma_locations_phenotypes, # deprecated
        ]
        phenotypes = []
        for func in phenotype_functions:
            phenotypes.extend(func(cell_specimen))
        return phenotypes

    def make_config(self):
        # have to call Config here because transgenic lines doesn't exist
        self.config = Config(name=self.name,
                             imports=[f'NIFRAW:{self.branch}/ttl/generated/allen-transgenic-lines.ttl'],
                             prefixes=self.prefixes,
                             branch=self.branch,
                             sources=tuple(),  # TODO insert the link to the query...
                             source_file=relative_path(
                                 __file__,
                                 no_wd_value=__file__))

    def build_neurons(self):
        instances = []
        dids = []
        done = {}
        for cell_specimen in self.neuron_data:
            sid = cell_specimen['id']
            tid = OntId(f'npokb:{ts_mapping[sid]}') if sid in ts_mapping else None

            phens = self.build_phenotypes(cell_specimen)
            cell = Cell(*phens)  # use the simple repr to id cells and get the tid
            if cell not in done:
                if tid is not None:
                    done[cell] = tid
            else:
                tid = done[cell]

            neuron = NeuronACT(*phens, id_=tid)
            did = AIBSSPEC[str(sid)]
            neuron.add_objects(ilxtr.hasNamedIndividual, did)
            dids.append(did)
            instances.append((did, rdf.type, owl.NamedIndividual))
            instances.append((did, rdf.type, neuron.identifier))

        print(sorted(self.tag_names))
        NeuronACT.write()
        NeuronACT.write_python()
        self.build_instances(instances, dids)

    def build_instances(self, instances, dids):
        folder = Path(self.config.out_graph_path()).parent
        # WOW do I need to implement the new/better way of
        # managing writing collections of neurons to graphs
        neuron_uri = next(NeuronACT.out_graph[:rdf.type:owl.Ontology])
        name = 'allen-cell-instances.ttl'
        base, _ = neuron_uri.rsplit('/', 1)
        uri = rdflib.URIRef(base + '/' + name)
        metadata = ((uri, rdf.type, owl.Ontology),)
        instance_graph = OntGraph(path=folder / name)
        instance_graph.bind('AIBSSPEC', AIBSSPEC)
        instance_graph.bind('npokb', npokb)
        [instance_graph.add(t) for t in metadata]
        [instance_graph.add(t) for t in instances]
        [instance_graph.add(t) for t in allDifferent(None, distinctMembers(*dids))]
        instance_graph.write()

    def build_transgenic_lines(self):
        """
        init class     |  "transgenic_line_source_name":"stock_number" a Class
        add superClass |  rdfs:subClassOf ilxtr:transgenicLine
        add *order*    |  ilxtr:useObjectProperty ilxtr:<order>
        add name       |  rdfs:label "name"
        add def        |  definition: "description"
        add transtype  |  rdfs:hasTransgenicType "transgenic_line_type_name"
        """

        triples = []
        for cell_specimen in self.neuron_data:
            for tl in cell_specimen['donor']['transgenic_lines']:
                _id = tl['stock_number'] if tl['stock_number'] else tl['id']
                prefix = tl['transgenic_line_source_name']
                line_type = tl['transgenic_line_type_name']
                if line_type == 'driver' and 'CreERT2' in tl['name']:
                    line_type = 'inducibleDriver'

                if prefix not in ['JAX', 'MMRRC', 'AIBS']:
                    print(tc.red('WARNING:'), 'unknown prefix', prefix, json.dumps(tl, indent=4))
                    continue
                elif prefix == 'AIBS':
                    prefix = 'AIBSTL'

                _class = self.ns[prefix][str(_id)]
                triples.append((_class, rdf.type, owl.Class))
                triples.append((_class, rdfs.label, rdflib.Literal(tl['name'])))
                triples.append((_class, definition, rdflib.Literal(tl['description'])))
                triples.append((_class, rdfs.subClassOf, ilxtr.transgenicLine))
                triples.append((_class, ilxtr.hasTransgenicType, ilxtr[line_type + 'Line']))

        # TODO aspects.ttl?
        transgenic_lines = simpleOnt(filename='allen-transgenic-lines',
                                     local_base=graphBase.local_base,
                                     path='ttl/generated/',
                                     prefixes=self.prefixes,
                                     triples=triples,
                                     comment='Allen transgenic lines for cell types',
                                     branch=self.branch,
                                     calling__file__=__file__,)

        transgenic_lines._graph.write()


def main(args={o.name:o.value for o in parse_defaults(__doc__)}, run=True):
    #print(args)
    if not args['--refresh'] and args['--input'] and Path(args['--input']).exists():
        with open(args['--input'], 'rt') as f:
            input = json.load(f)
    else:
        response = requests.get('http://api.brain-map.org/api/v2/data/query.json?criteria='
                                'model::Specimen,rma::criteria,[is_cell_specimen$eq%27true%27],'
                                'products[name$eq%27Mouse%20Cell%20Types%27],'
                                'rma::include,structure,donor(transgenic_lines),'
                                'specimen_tags,cell_soma_locations,rma::options[num_rows$eqall]')
        input = response.json()['msg']
        with open(args['--input'], 'wt') as f:
            json.dump(input, f, indent=4)

    input = sorted(input, key=lambda d: d['id'])  # ensure type specimens show up first
    act = AllenCellTypes(input, args['--output'])
    act.make_config()
    if __name__ == '__main__' or run:
        act.build_transgenic_lines()
        act.build_neurons()

    return act.config,


if __name__ == '__main__':
    args = docopt(__doc__, version='0.0.4')
    main(args)



import augpathlib as aug
from pyontutils.core import OntGraph, OntResPath
from pyontutils.config import auth
from pyontutils.namespaces import makePrefixes


#import json
def path_json(p):
    with open(p, 'rt') as f:
        return json.load(f)


def asObj(r, name=None):
    # SIGH functions and errors do not compose well at all
    d = {h: getattr(r, h)().value for h in r.header}
    errors = [v for v in d.values() if len(v) != len(v.strip())]
    if errors:
        msg = f'BEEP BOOP WHITESPACE VIOLATION IN {name}: {errors!r}'
        raise ValueError(msg)

    return d


def _genlabel(oid):
    x = (oid.rsplit("-", 3)[-3]
         if oid.endswith('-prime')
         else (oid.rsplit("-", 3)[-3] if 'unbranched' in oid else oid.rsplit("-", 2)[-2]))
    l = (oid.rsplit("-", 3)[-3]
         if oid.endswith('-prime')
         else (' '.join(oid.rsplit("-", 3)[-3:-1]) if 'unbranched' in oid else oid.rsplit("-", 2)[-2]))
    t = (oid.rsplit("-", 2)[-2] + "'"
         if oid.endswith('-prime')
         else (oid.rsplit("-", 1)[-1] if 'unbranched' in oid else oid.rsplit("-", 1)[-1]))

    # XXX manual fix for consistency
    if x == 'keast':
        x = 'kblad'
    if l == 'keast':
        l = 'kblad'

    label = f'neuron type {l} {t}'
    return label, x


from pprint import pprint
#from pyontutils.sheets import Sheet
from sparcur.sheets import Sheet  # FIXME move the caching functionality to pyontutils.sheets
# XXX for now if you make a change you can e.g. insts[0].fetch(_refresh_cache=True)

Sheet._do_cache = True

amr = aug.LocalPath("~/git/apinatomy-models").expanduser()
skip = 'too-map', 'dev-layout-conn', 'scaffold-test', 'vagus-nerve', 'wbrcm'
models = [_.name for _ in (amr / 'models').children if _.name not in skip and (_ / f'source/{_.name}.xlsx').exists()]

# might as well just pull the sheets directly at this point
#_json_paths = {name: amr / 'models' / name / 'source' / f'{name}.json' for name in models}
#jsons = {n: path_json(p) for n, p in _json_paths.items()}

classes = []
main_classes = []  # XXX because abbreviation is not included in the json !!!
refs_classes = []
lcnv_classes = []
for model in models:
    c = type(f'{model.replace("-", "_")}groups', (Sheet,), {'name': f'{model}', 'sheet_name': 'groups'})
    classes.append(c)
    c = type(f'{model.replace("-", "_")}main', (Sheet,), {'name': f'{model}', 'sheet_name': 'main'})
    main_classes.append(c)
    c = type(f'{model.replace("-", "_")}references', (Sheet,), {'name': f'{model}', 'sheet_name': 'references'})
    refs_classes.append(c)
    c = type(f'{model.replace("-", "_")}localConventions', (Sheet,), {'name': f'{model}', 'sheet_name': 'localConventions'})
    lcnv_classes.append(c)

insts = [c(fetch=False) for c in classes]
main_insts = [c(fetch=False) for c in main_classes]
refs_insts = [c(fetch=False) for c in refs_classes]
lcnv_insts = [c(fetch=False) for c in lcnv_classes]

#print(block_lines)
from neurondm.lang import Phenotype, Neuron, Config, OntId
#i = insts[0]
#pprint(dir(i.row_object(0)))

#pprint([i.values for i in insts])

#para_pre =  'ilxtr:neuron-phenotype-para-pre'
#para_post = 'ilxtr:neuron-phenotype-para-post'
#sym_pre =   'ilxtr:neuron-phenotype-sym-pre'
#sym_post =  'ilxtr:neuron-phenotype-sym-post'

para = Phenotype('ilxtr:ParasympatheticPhenotype', 'ilxtr:hasAnatomicalSystemPhenotype')
sym = Phenotype('ilxtr:SympatheticPhenotype', 'ilxtr:hasAnatomicalSystemPhenotype')
pre = Phenotype('ilxtr:PreGanglionicPhenotype', 'ilxtr:hasAnatomicalSystemPhenotype')
post = Phenotype('ilxtr:PostGanglionicPhenotype', 'ilxtr:hasAnatomicalSystemPhenotype')

sens = Phenotype('ilxtr:SensoryPhenotype', 'ilxtr:hasCircuitRolePhenotype')
motor = Phenotype('ilxtr:MotorPhenotype', 'ilxtr:hasCircuitRolePhenotype')
intrin = Phenotype('ilxtr:IntrinsicPhenotype', 'ilxtr:hasCircuitRolePhenotype')

excite = Phenotype('ILX:0104003', 'ilxtr:hasFunctionalCircuitRolePhenotype')
inhib = Phenotype('ILX:0105486', 'ilxtr:hasFunctionalCircuitRolePhenotype')

asc = Phenotype('ilxtr:SpinalCordAscendingProjectionPhenotype', 'ilxtr:hasProjectionPhenotype')
desc = Phenotype('ilxtr:SpinalCordDescendingProjectionPhenotype', 'ilxtr:hasProjectionPhenotype')
intfug = Phenotype('ilxtr:IntestinoFugalProjectionPhenotype', 'ilxtr:hasProjectionPhenotype')

enteric = Phenotype('ilxtr:EntericPhenotype')  # probably not just partOf some enteric nervous system ... though maybe given developmental origin
nos = Phenotype('TEMPIND:Nos1', 'ilxtr:hasMolecularPhenotype')

mapping = {
    'sympathetic': sym,
    'parasympathetic': para,
    'pre-ganglionic': pre,
    'post-ganglionic': post,
    'sensory': sens,
    'motor': motor,
    'local': intrin,
    'interneuron': intrin,
    'excitatory': excite,
    'inhibitory': inhib,

    'intestinofugal': intfug,
    'enteric': enteric,
    'descending': desc,
    'ascending': asc,
    'NOS': nos,
}

bad_bits = {'neuron', '', 'first', 'order', 'circuit', 'lower', 'somatic', 'Vagal'}
bad_sets = (  # from the school of maximum wat
    {'inhibitory', 'motor'},
    # inhibitory motor in the context of the stomach means nitric
    # oxide releaseing, motor usually implies actuating ach, we need
    # to clarify the definition because neuro/myo/modulatory release
    # is usually not what we mean in other systems
    {'primary', 'afferent', 'intrinsic'},
    {'visceral', 'sensory'},  # I think this one is just wrong
)

exact_sets = {
frozenset(('')): 'hello',
}


def normalize_type(string):
    raw_bits = set(string.split(' '))
    bits = raw_bits - bad_bits
    if bits in bad_sets:
        return set()
    id_bits = set([mapping[b] if b in mapping else b for b in bits
                   if b in mapping])
    missed = [b for b in bits if b not in mapping]
    return id_bits, missed


def parse_refs(refs):
    return [_ for _ in [r.strip() for r in refs.split(',')] if _]


_id_oid_type_refs = [
    (r.id().value,
     r.ontologyterms().value if hasattr(r, 'ontologyterms') else f'ERROR {inst}',
     r.neurontypes().value if hasattr(r, 'neurontypes') else f'ERROR {inst}',
     r.references().value if hasattr(r, 'references') else f'ERROR {inst}',)
    for inst in insts if hasattr(inst, '_values') or inst.fetch() for r in inst.rows()[1:]]
id_oid_type_refs = [(i, o, t, refs) for i, o, t, refs in _id_oid_type_refs if o]
#pprint(id_oid_type)
raw_types = sorted(set([t for i, o, t, refs in id_oid_type_refs]))
#pprint(raw_types)
#pprint([(t, normalize_type(t)) for t in raw_types])
bagged_missed = [(normalize_type(t), t, o, parse_refs(refs)) for i, o, t, refs in id_oid_type_refs]
bagged = [(a, c, d, b, r) for ((a, b), c, d, r) in bagged_missed]
#pprint(bagged, width=240)


id_sn = {r.id().value: r.abbreviation().value if hasattr(r, 'abbreviation') else 'MISSING ???'
         for i in main_insts if hasattr(i, '_values') or i.fetch() for r in i.rows()[1:]}
sn_id = {s:i for i, s in id_sn.items()}

# model level refs
mrefs = {i.name: [asObj(r, i.name) for r in i.rows()[1:]] for i in refs_insts if hasattr(i, '_values') or i.fetch()}

lcnv = {i.name: {r.prefix().value if hasattr(r, 'prefix') else 'AAAA':  # r.namespace().value
                 r.namespace().value if hasattr(r, 'namespace') else 'BBBB'
                 for r in i.rows()[1:]} for i in lcnv_insts if hasattr(i, '_values') or i.fetch()}

neuron_classes = {
    sn: type(f'Neuron{sn.capitalize()}', (Neuron,), {'owlClass': f'ilxtr:Neuron{sn.capitalize()}',
                                                     '_model_refs': mrefs[sn_id[sn]],
                                                     '_model_id': sn_id[sn],
                                                     })
    for sn in [('kblad' if _ == 'keast' else
                ('bolew' if _ == 'unbranched' else _))
               for _ in
               set(oid.rsplit("-", 3)[-3]
                   if oid.endswith('-prime')
                   else oid.rsplit("-", 2)[-2]
                   for _, oid, _, _ in id_oid_type_refs)]
}

# TODO add citations to parent classes or individual neurons, TODO probably also link to the apinatomy model itself
# TODO add links to svgs to the Neuron{Model} classes
svg_template = (
    'https://raw.githubusercontent.com/open-physiology/apinatomy-models/'
    'master/models/{id_model}/docs/{id_model}.svg')


def dophen(oid, bag, refs):
    label, x = _genlabel(oid)
    neuron_class = neuron_classes[x]
    # FIXME this doesn't work because entailed phenotypes don't work
    # with the correct logic because they are still acting like ec
    # phenotypes which replace eachother, ent neurons should stack
    # so for now we just stick the neuron id itself in and don't worry about it
    u = OntId(oid).u
    up = OntId(oid).u.replace('neuron-type', 'neuron-phenotype')
    nc = neuron_class(Phenotype(up, 'ilxtr:hasClassificationPhenotype'), *(p.asEntailed() for p in bag),
                      #label=label,
                      id_=u)
    nc._refs = refs
    return nc


def citations(ncs, neus):
    p = OntId('ilxtr:literatureCitation').u
    #hsa = OntId('ilxtr:hasSourceArtifact').u  # XXX doesn't quite match usage because it is pseudo ontological
    source = OntId('dc:source').u  # more reasonable given the complexity of the process
    for things, attr, idf in ((ncs, '_model_refs', lambda c: OntId(c.owlClass).u),
                              (neus, '_refs', lambda c: c.identifier)):
        for thing in things:
            s = idf(thing)

            if attr == '_model_refs':
                yield s, source, OntId(svg_template.format(id_model=thing._model_id)).u

            for ref in getattr(thing, attr):
                if isinstance(ref, dict):
                    ref = ref['id']  # XXX FIXME WARNING EVIL

                try:
                    o = OntId(ref).u
                except Exception as e:
                    raise ValueError(s) from e

                yield s, p, o


# XXX TODO need to ensure that the type phenotype axioms so that the model superclass is included
config = Config('apinat-pops-more')
neus = [dophen(oid, bag, refs) for bag, _, oid, _, refs in bagged]
#test = list(citations(neuron_classes.values(), neus))
pprint(neus)
#sys.modules['__main__'].__file__ = __file__
#import linecache
#linecache.cache[__file__] = size, mtime, lines, fullname
config.write()
config.write_python()

# cleanup
ogp = config._written_graph.path  # XXX cannot trust config.out_graph_path() due to failover if path does not exist in graphBase >_< hooray
og = OntGraph(path=ogp)
og.parse()
ng = OntGraph().populate_from_triples((t for t in og if 'able' not in t[1] and 'abel' not in t[1]))
ng.populate_from(citations(neuron_classes.values(), neus))
ng.namespace_manager.populate_from(og)
ng.namespace_manager.populate_from(makePrefixes('PMID', 'doi', 'dc'))
ng.write(ogp)

if False:
    pprint([(a, b, c, d) for a, b, c, d in bagged if not a])
    pprint([(d, b, c) for a, b, c, d in bagged if d])



#!/usr/bin/env python3

import inspect
import neurondm
from pyontutils.utils import TermColors as tc
from neurondm import *
from neurondm import phenotype_namespaces as phns
try:
    breakpoint
except NameError:
    from IPython import embed as breakpoint

config = Config('neurons-example-script')

def printe(*args, **kwargs):
    print(*(tc.blue(str(a)) for a in args), **kwargs)

def printer(*args, **kwargs):
    printe(*(tc.red(repr(a)) for a in args), **kwargs)


def messup(outer_n):
    print('neurons defined outside local scope')
    print(repr(outer_n))
    print(outer_n)
    setLocalNames(phns.BBP)  # testing to make sure we get an error if we call this in a function
    # there is no error and the names stick around after the function returns
    print(repr(outer_n), outer_n)
    print('neurons defined in local scope')
    n = Neuron(brain, Phenotype('PR:000013502'))
    print(repr(n))

# namespace example
with phns.Layers:
    L23
    try: brain
    except NameError: printe('brain fails as expected')
    with phns.Regions:
        L23
        brain
        Neuron(L23, brain)
        try: Mouse
        except NameError: printe('Mouse fails as expected')
        with phns.Species:
            Neuron(Mouse, L23, CTX)
        try: Mouse
        except NameError: printe('Mouse fails as expected')
    try: brain
    except NameError: printe('brain fails as expected')
    L1
try: L1
except NameError: printe('L1 fails as expected')

with phns.Layers, phns.Regions, phns.Species:
    try:
        with phns.Test:
            pass
    except ValueError as e:
        printe('Test namespace has conflicts and fails as expected.', e)
    Neuron(Rat, L23, CTX)

# context example
with phns.BBP:
    myFirstNeuron = Neuron(Rat, CTX, L23, PV)
    mySecondNeuron = Neuron(Mouse, CA1, SO, PV)
    with myFirstNeuron:  # TODO __add__ for neurons that fails on disjointness
        myThirdNeuron = Neuron(LBC)
        with mySecondNeuron:  # contexts are mutually exclusive even when nested WARNING MAY CHANGE!
            myFourthNeuron = Neuron(LBC)
        with myThirdNeuron:  # use a neuron defined in a context as context to simulate nesting
            try: myFifthNeuron = Neuron(SBC)  # trivia: this neuron can't actually exist as defined
            except TypeError: printe('Neuron(SBC) fails as expected in context of LBC')
            myFifthNeuron = Neuron(SOM)  # trivia: this neuron can't actually exist as defined
    mySixthNeuron = Neuron(brain)
    print('first ', repr(myFirstNeuron))
    print('second', repr(mySecondNeuron))
    print('third ', repr(myThirdNeuron))
    print('fourth', repr(myFourthNeuron))
    print('fifth ', repr(myFifthNeuron))
    print('sixth ', repr(mySixthNeuron))

    do_not_expect = Neuron(Rat)
    with Neuron(brain) as n0, Neuron(Rat) as n1:  # they appear to combine but observer that n1 != Neuron(Rat)
        n2 = Neuron(PV)
        assert n1 != do_not_expect
        printer(n0, n1, n2)

    expect = Neuron(brain, Rat, PV)
    with Neuron(SOM) as n3, n1:  # last one wins when setting multiple contexts
        n4 = Neuron(PV)
        assert n4 == expect
        printer(n3, n1, n4)

# addition
try: myFirstNeuron + mySecondNeuron
except TypeError: printe('myFirstNeuron + mySecondNeuron fails as expected')
with phns.BBP:
    sumn = Neuron(S1, SOM) + Neuron(Mouse, PV)
    print('sumn a neuron', repr(sumn))
    nn = sum((sumn, Neuron(NGC)))
    print('sumn a neuron again', repr(nn))

setLocalNames(phns.BBP)
setLocalContext(Phenotype('NCBITaxon:10090', ilxtr.hasInstanceInTaxon))
Neuron(Phenotype('UBERON:0001950', ilxtr.hasSomaLocatedIn, label='neocortex'))
Neuron(brain, Phenotype('PR:000013502'))
Neuron(Phenotype('UBERON:0000955'), Phenotype('CHEBI:18243'))
Neuron(Phenotype('UBERON:0001950', ilxtr.hasSomaLocatedIn))
Neuron(Phenotype('UBERON:0000955'), Phenotype('CHEBI:18243'), Phenotype('PR:000013502'))

def inner():
    Neuron(SOM, Phenotype('PR:000013502'))
inner()

#resetLocalNames()  # works as expected at the top level
#resetLocalNames(globals())  # works as expected
pv = Neuron(brain, Phenotype('PR:000013502'))

setLocalNames()
messup(pv)  # the localNames call inside here persists
print('testing printing pv after localNames is called inside messup')
print(repr(pv))

print(config.neurons())

if __name__ == '__main__':
    breakpoint()

# XXX these have to be called inside this module or the state persists in graphBase FIXME
resetLocalNames()
setLocalContext()



#!/usr/bin/env python3

from pathlib import Path
import rdflib
from pyontutils.utils import rowParse, relative_path
from pyontutils.namespaces import ilxtr
from neurondm.lang import *
from neurondm.phenotype_namespaces import BBP


class NeuronMarkram2015(NeuronEBM):
    owlClass = ilxtr.NeuronMarkram2015
    shortname = 'Markram2015'


class table1(rowParse):
    citation = 'Markram et al Cell 2015'
    pmid = 'PMID:26451489'
    _sep = '|'

    def __init__(self, *args, **kwargs):
        self._id_order = []
        with BBP:
            self._context = Neuron(Rat, S1, Interneuron, GABA, id_=OntId('npokb:115'))
            super().__init__(*args, **kwargs)

    def Morphological_type(self, value):
        syn, abrv = value.split(' (')
        syn = syn.strip()
        abrv = abrv.rstrip(')').strip()
        # print((syn, abrv))
        self._id_order.append((0, 'm', abrv))
        self._mtype = BBP[abrv]

        self._m_syn = syn
        self._m_abrev = abrv
        return self._mtype

    def Other_morphological_classifications(self, value):
        return  #  skipping all of this for now due to bitufted nonsense etc
        values = value.split(self._sep)
        output = []
        callbacks = []

        for v in values:
            if '/' in v:
                prefix, a_b = v.split(' ')
                a, b = a_b.split('/')
                #output.append(prefix + ' ' + a)
                #output.append(prefix + ' ' + b)
            else:
                prefix = v.rstrip('cell').strip()
                #output.append(v)

            label = prefix + ' phenotype'
            output.append(label)

        for v in output:
            self.graph.add_trip(self._morpho_parent_id, 'NIFRID:synonym', v)

    def Predominantly_expressed_Ca2_binding_proteins_and_peptides(self, value):
        p_edge = ilxtr.hasExpressionPhenotype
        with BBP:
            p_map = {
                'CB':CB,
                'PV':PV,
                'CR':CR,
                'NPY':NPY,
                'VIP':VIP,
                'SOM':SOM,
            }

        NEGATIVE = False
        POSITIVE = True  # FIXME this requires more processing prior to dispatch...
        e_edge = ''
        e_map = {
            '-':NEGATIVE,
            '+':POSITIVE,
            '++':POSITIVE,
            '+++':POSITIVE,
        }
        NONE = 0
        LOW = 1
        MED = 2
        HIGH = 3
        s_map = {
            '-':NONE,
            '+':LOW,
            '++':MED,
            '+++':HIGH,
        }

        values = value.split(self._sep)
        self._moltypes = []
        for v in values:
            abrv, score_paren = v.split(' (')
            score = score_paren.rstrip(')')
            molecule = BBP[abrv] #p_map[abrv]
            exists = e_map[score]
            score = s_map[score]
            if exists:
                self._id_order.append((0, 'ex', abrv))
                self._moltypes.append(molecule)
            else:
                self._id_order.append((1, 'ex', abrv))
                self._moltypes.append(NegPhenotype(molecule))

        return self._moltypes

    def Electrical_types(self, value):  # FIXME these are mutually exclusive types, so they force the creation of subClasses so we can't apply?
        e_map = {
            'b':b,
            'c':c,
            'd':d,
        }
        l_map = {
            'AC':AC,
            'NAC':NAC,
            'STUT':STUT,
            'IR':IR,
        }

        values = value.split(self._sep)
        self._e_id_order = []
        self._etypes = []
        for v in values:
            early_late, score_pct_paren = v.split(' (')
            score = int(score_pct_paren.rstrip('%)'))
            e_name, l_name = early_late[0], early_late[1:]
            #early, late = e_map[e_name], l_map[l_name]
            early, late = BBP[e_name], BBP[l_name]
            lpe = LogicalPhenotype(AND, early, late)
            self._e_id_order.append((4, '', ((0, 'el', 'i' + e_name), (0, 'el', 's' + l_name.lower()))))
            self._etypes.append(lpe)

        return self._etypes

    def Other_electrical_classifications(self, value):
        valid_mappings = {'Fast spiking':FS,
                          'Non-fast spiking':NegPhenotype(FS),  # only in this very limited context
                          'Regular spiking non-pyramidal':RSNP}

        values = value.split(self._sep)
        self._other_etypes = []
        for v in values:
            if v in valid_mappings:
                self._id_order.append((1, 'el', 'Fast spiking')
                                      if v.startswith('Non-') else
                                      (0, 'el', v))
                self._other_etypes.append(valid_mappings[v])

        return self._other_etypes

    def _row_post(self):
        mio = sorted([o for o in self._id_order if o[1] == 'm'] +
                     [tuple((8 if o[0] == 0 else 8.5,) + o[1:])
                      for o in self._id_order if o[1] == 'ex'])
        with self._context:
            n = NeuronMarkram2015(self._mtype,
                                  *[m.asEntailed()
                                    for m in self._moltypes],
                                  label=self._m_syn)
            n._ido = mio
            n.abbrevs = [rdflib.Literal(self._m_abrev)]

            for etype, eio in zip(self._etypes, self._e_id_order):
                io = sorted([*self._id_order, eio])
                n = NeuronMarkram2015(
                    etype, self._mtype, *self._other_etypes, *self._moltypes)
                n._ido = io

        self._id_order = []

    def _end(self):
        def key(n):
            return n._ido

        def sgid(n, i):
            index = 59 + i
            oid = n.id_
            n.id_ = OntId(f'npokb:{index}').u

        need_id = sorted([n for n in Neuron.existing_pes if hasattr(n, '_ido')], key=key)
        [sgid(n, i) for i, n in enumerate(need_id)]


def main():
    import csv
    from neurondm.core import auth
    with open(auth.get_path('resources') / '26451489 table 1.csv', 'rt') as f:
        rows = [list(r) for r in zip(*csv.reader(f))]

    config = Config('blank')  # protect from inheriting state if main is imported XXX SIGH
    table1(rows)
    ep = Neuron.existing_pes

    config = Config('markram-2015',
                    source_file=relative_path(__file__, no_wd_value=__file__))

    for _n in ep:
        _n._sighed = False

    [n._sigh() for n in ep]
    #[graphBase.out_graph.add((n.id_, ilxtr.hasTemporaryId, n.temp_id)) for n in ep]
    del NeuronMarkram2015._runonce
    NeuronMarkram2015.__new__(NeuronMarkram2015)
    graphBase.out_graph.add((NeuronMarkram2015.owlClass,
                             ilxtr.modelSource,
                             OntId('https://doi.org/10.1016/j.cell.2015.09.029').u))

    config.write()
    config.write_python()
    return config,


__globals__ = globals()  # fuck you python

if __name__ == '__main__':
    main()
else:
    __globals__ = globals()
